import logging, os, sys

from ..appengine_info import DEBUG

assert DEBUG
creds = os.getenv('GOOGLE_APPLICATION_CREDENTIALS')
assert not creds or creds.endswith('fake_user_account.json')

logging.getLogger('chardet').setLevel(logging.INFO)
logging.getLogger('google.cloud').setLevel(logging.INFO)
logging.getLogger('lexrpc').setLevel(logging.INFO)
logging.getLogger('negotiator').setLevel(logging.WARNING)
logging.getLogger('urllib3').setLevel(logging.INFO)

# Piggyback on unittest's -v and -q flags to show/hide logging.
logging.basicConfig()
if '-v' in sys.argv:
  logging.getLogger().setLevel(logging.DEBUG)
else:
  # used to be: elif 'discover' in sys.argv or '-q' in sys.argv or '--quiet' in sys.argv:
  # dropped that to suppress logging when running full single test files
  logging.disable(logging.CRITICAL + 1)



"""A handler that serves all app logs for an App Engine HTTP request.

StackDriver Logging API:
https://cloud.google.com/logging/docs/apis
"""
import calendar
from datetime import datetime, timedelta, timezone
import html
import logging
import re
import time
import urllib.request, urllib.parse, urllib.error

from flask import request
from google.cloud import ndb
from google.cloud.logging import Client
from oauth_dropins.webutil.util import json_dumps, json_loads

from .appengine_info import APP_ID
from . import flask_util, util
from .flask_util import error

logger = logging.getLogger(__name__)

LEVELS = {
  logging.DEBUG:    'D',
  logging.INFO:     'I',
  logging.WARNING:  'W',
  logging.ERROR:    'E',
  logging.CRITICAL: 'F',
}

CACHE_TIME = timedelta(days=1)
MAX_LOG_AGE = timedelta(days=30)
# App Engine's launch, roughly
MIN_START_TIME = time.mktime(datetime(2008, 4, 1, tzinfo=timezone.utc).timetuple())
MAX_START_TIME = time.mktime(datetime(2099, 1, 1, tzinfo=timezone.utc).timetuple())

SANITIZE_RE = re.compile(r"""
  ((?:access|api|oauth)?[ _]?
   (?:code|consumer_key|consumer_secret|nonce|secret|signature|token|verifier)
     (?:u?['"])?
   (?:=|:|\ |,\ |%3D)\ *
     (?:u?['"])?
  )
  [^ &='"]+
""", flags=re.VERBOSE | re.IGNORECASE)


def sanitize(msg):
  """Sanitizes access tokens and Authorization headers."""
  return SANITIZE_RE.sub(r'\1...', msg)


def url(when, key, **params):
  """Returns the relative URL (no scheme or host) to a log page.

  Args:
    when: datetime
    key: ndb.Key or str
    params: included as query params, eg module, path
  """
  assert 'start_time' not in params and 'key' not in params, params

  if isinstance(params.get('path'), (list, tuple)):
    for path in params['path']:
      assert ',' not in path, path
    params['path'] = ','.join(params['path'])

  params.update({
    'start_time': calendar.timegm(when.utctimetuple()),
    'key': key.urlsafe().decode() if isinstance(key, ndb.Key) else key,
  })
  return 'log?' + urllib.parse.urlencode(params)


def maybe_link(when, key, time_class='dt-updated', link_class='', **params):
  """Returns an HTML snippet with a timestamp and maybe a log page link.

  Example:

  <a href="/log?start_time=1513904267&key=aglz..." class="u-bridgy-log">
    <time class="dt-updated" datetime="2017-12-22T00:57:47.222060"
            title="Fri Dec 22 00:57:47 2017">
      3 days ago
    </time>
  </a>

  The <a> tag is only included if the timestamp is 30 days old or less, since
  Stackdriver's basic tier doesn't store logs older than that:
    https://cloud.google.com/monitoring/accounts/tiers#logs_ingestion
    https://github.com/snarfed/bridgy/issues/767

  Args:
    when: datetime
    key: ndb.Key or str
    time_class: string, optional class value for the <time> tag
    link_class: string, optional class value for the <a> tag (if generated)
    params: dict {string: string}, query params to include in the link URL,
      eg module, path

  Returns: string HTML
  """
  # always show time zone. assume naive timestamps are UTC.
  if when.tzinfo is None:
    when = when.replace(tzinfo=timezone.utc)

  now = util.now(tz=when.tzinfo)

  time = f'<time class="{time_class}" datetime="{when.isoformat()}" title="{when.ctime()} {when.tzname()}">{util.naturaltime(when, when=now)}</time>'

  if now > when > now - MAX_LOG_AGE:
    return f'<a class="{link_class}" href="/{url(when, key, **params)}">{time}</a>'

  return time


# datastore string keys are url-safe-base64 of, say, at least 32(ish) chars.
# https://cloud.google.com/appengine/docs/python/ndb/keyclass#Key_urlsafe
# http://tools.ietf.org/html/rfc3548.html#section-4
BASE64 = 'A-Za-z0-9-_='
DATASTORE_KEY_RE = re.compile("([^%s])(([%s]{8})[%s]{24,})([^%s])" % ((BASE64,) * 4))

def linkify_datastore_keys(msg):
  """Converts string datastore keys to links to the admin console viewer."""
  def linkify_key(match):
    try:
      # Useful for logging, but also causes false positives in the search, we
      # find and use log requests instead of real requests.
      # logger.debug(f'Linkifying datastore key: {'match.group(2)}')
      key = ndb.Key(urlsafe=match.group(2))
      tokens = [(kind, f"{'id' if isinstance(id, int) else 'name'}:{id}")
                for kind, id in key.pairs()]
      key_str = '0/|' + '|'.join(f'{len(kind)}/{kind}|{len(id)}/{id}'
                                 for kind, id in tokens)
      key_quoted = urllib.parse.quote(urllib.parse.quote(key_str, safe=''), safe='')
      html = f"{match.group(1)}<a title='{match.group(2)}' href='https://console.cloud.google.com/datastore/entities;kind={key.kind()};ns=__$DEFAULT$__/edit;key={key_quoted}?project={APP_ID}'>{match.group(3)}...</a>{match.group(4)}"
      logger.debug(f'Returning {html}')
      return html
    except BaseException:
      # logger.debug("Couldn't linkify candidate datastore key.")   # too noisy
      return match.group(0)

  return DATASTORE_KEY_RE.sub(linkify_key, msg)


def log(module=None, path=None):
    """Flask view that searches for and renders app logs for an HTTP request.

    URL parameters:
      start_time: float, seconds since the epoch
      key: string that should appear in the first app log

    Install with:
      app.add_url_rule('/log', view_func=logs.log)

    Or:
      @app.get('/log')
      @cache.cached(600)
      def log():
        return logs.log()

    Args:
      module: str, App Engine module to search. Defaults to all.
      path: string or sequence of strings, optional HTTP request path(s) to
        limit logs to.

    Returns:
      (string response body, dict headers) Flask response
    """
    if not module:
      module = request.values.get('module')
    if not path:
      path = request.values.get('path')

    start_time = flask_util.get_required_param('start_time')
    if not util.is_float(start_time):
      return error(f"Couldn't convert start_time to float: {start_time!r}")

    start_time = float(start_time)
    if start_time < MIN_START_TIME:
      return error(f'start_time must be >= {MIN_START_TIME}')
    elif start_time > MAX_START_TIME:
      return error(f'start_time must be <= {MAX_START_TIME}')

    client = Client()
    project = f'projects/{APP_ID}'
    key = urllib.parse.unquote_plus(flask_util.get_required_param('key'))

    # first, find the individual log message to get the trace id
    utcfromtimestamp = datetime.utcfromtimestamp
    timestamp_filter = (
      f"timestamp>=\"{utcfromtimestamp(start_time - 60).isoformat() + 'Z'}\" "
      f"timestamp<=\"{utcfromtimestamp(start_time + 120).isoformat() + 'Z'}\"")
    query = f'logName="{project}/logs/python" textPayload:"{key}" {timestamp_filter}'
    if module:
      query += f' resource.labels.module_id="{module}"'
    if path:
      or_paths = ' OR '.join(f'"{path}"' for path in path.split(','))
      query += f' httpRequest.requestUrl:({or_paths})'

    logger.info(f'Searching logs with: {query}')
    try:
      # https://googleapis.dev/python/logging/latest/client.html#google.cloud.logging_v2.client.Client.list_entries
      log = next(iter(client.list_entries(filter_=query, page_size=1)))
    except StopIteration:
      logger.info('No log found!')
      return 'No log found!', 404

    logger.info(f'Got insert id {log.insert_id} trace {log.trace}')

    # now, print all logs with that trace
    resp = """\
<html>
<body style="font-family: monospace; white-space: pre">
"""

    query = f'logName="{project}/logs/python" trace="{log.trace}" resource.type="gae_app" {timestamp_filter}'
    logger.info(f'Searching logs with: {query}')

    # sanitize and render each line
    for log in client.list_entries(filter_=query, page_size=1000):
      # payload is a union that can be string, JSON, or protobuf
      # https://cloud.google.com/logging/docs/reference/v2/rest/v2/LogEntry#FIELDS.oneof_payload
      msg = log.payload
      if not msg:
        continue
      elif isinstance(msg, (dict, list)):
        msg = json_dumps(msg, indent=2)
      else:
        msg = str(msg)

      msg = linkify_datastore_keys(util.linkify(html.escape(
        msg if msg.startswith('Created by this poll:') else sanitize(msg),
        quote=False)))
      resp += '%s %s %s<br />' % (
        log.severity[0], log.timestamp, msg.replace('\n', '<br />'))

    resp += '</body>\n</html>'
    return resp, {'Content-Type': 'text/html; charset=utf-8'}



"""LinkedIn OAuth drop-in.

API docs:
https://www.linkedin.com/developers/
https://docs.microsoft.com/en-us/linkedin/consumer/integrations/self-serve/sign-in-with-linkedin
"""
import logging
import urllib.parse

from flask import request
from google.cloud import ndb

from . import views
from .models import BaseAuth
from .webutil import flask_util, util
from .webutil.util import json_dumps, json_loads

logger = logging.getLogger(__name__)

LINKEDIN_CLIENT_ID = util.read('linkedin_client_id')
LINKEDIN_CLIENT_SECRET = util.read('linkedin_client_secret')
# URL templates. Can't (easily) use urlencode() because I want to keep
# the %(...)s placeholders as is and fill them in later in code.
AUTH_CODE_URL = '&'.join((
  'https://www.linkedin.com/oauth/v2/authorization?'
  'response_type=code',
  'client_id=%(client_id)s',
  # https://docs.microsoft.com/en-us/linkedin/shared/integrations/people/profile-api?context=linkedin/consumer/context#permissions
  'scope=%(scope)s',
  # must be the same in the access token request
  'redirect_uri=%(redirect_uri)s',
  'state=%(state)s',
))
ACCESS_TOKEN_URL = 'https://www.linkedin.com/oauth/v2/accessToken'
API_PROFILE_URL = 'https://api.linkedin.com/v2/me'


class LinkedInAuth(BaseAuth):
  """An authenticated LinkedIn user.

  Provides methods that return information about this user and make OAuth-signed
  requests to the LinkedIn REST API. Stores OAuth credentials in the datastore.
  See models.BaseAuth for usage details.

  Implements get() but not urlopen() or api(). The key name is the ID (a URN).

  Note that LI access tokens can be over 500 chars (up to 1k!), so they need to
  be TextProperty instead of StringProperty.
  https://docs.microsoft.com/en-us/linkedin/shared/authentication/authorization-code-flow?context=linkedin/consumer/context#access-token-response
  """
  access_token_str = ndb.TextProperty(required=True)
  user_json = ndb.TextProperty()

  def site_name(self):
    return 'LinkedIn'

  def user_display_name(self):
    """Returns the user's first and last name.
    """
    def name(field):
      user = json_loads(self.user_json)
      loc = user.get(field, {}).get('localized', {})
      if loc:
          return loc.get('en_US') or loc.values()[0]
      return ''

    return f"{name('firstName')} {name('lastName')}"

  def access_token(self):
    """Returns the OAuth access token string.
    """
    return self.access_token_str

  def get(self, *args, **kwargs):
    """Wraps requests.get() and adds the Bearer token header.

    TODO: unify with github.py, medium.py.
    """
    return self._requests_call(util.requests_get, *args, **kwargs)

  def post(self, *args, **kwargs):
    """Wraps requests.post() and adds the Bearer token header.

    TODO: unify with github.py, medium.py.
    """
    return self._requests_call(util.requests_post, *args, **kwargs)

  def _requests_call(self, fn, *args, **kwargs):
    headers = kwargs.setdefault('headers', {})
    headers['Authorization'] = 'Bearer ' + self.access_token_str

    resp = fn(*args, **kwargs)

    try:
      resp.raise_for_status()
    except BaseException as e:
      util.interpret_http_exception(e)
      raise
    return resp


class Start(views.Start):
  """Starts LinkedIn auth. Requests an auth code and expects a redirect back.
  """
  NAME = 'linkedin'
  LABEL = 'LinkedIn'
  DEFAULT_SCOPE = 'r_liteprofile'

  def redirect_url(self, state=None):
    # assert state, 'LinkedIn OAuth 2 requires state parameter'
    assert LINKEDIN_CLIENT_ID and LINKEDIN_CLIENT_SECRET, \
      "Please fill in the linkedin_client_id and linkedin_client_secret files in your app's root directory."
    return AUTH_CODE_URL % {
      'client_id': LINKEDIN_CLIENT_ID,
      'redirect_uri': urllib.parse.quote_plus(self.to_url()),
      'state': urllib.parse.quote_plus(state or ''),
      'scope': self.scope,
    }

  @classmethod
  def button_html(cls, *args, **kwargs):
    return super(cls, cls).button_html(
      *args,
      input_style='background-color: #EEEEEE; padding: 5px; padding-top: 8px; padding-bottom: 2px',
      **kwargs)


class Callback(views.Callback):
  """The OAuth callback. Fetches an access token and stores it.
  """
  def dispatch_request(self):
    # handle errors
    error = request.values.get('error')
    desc = request.values.get('error_description')
    if error:
      # https://docs.microsoft.com/en-us/linkedin/shared/authentication/authorization-code-flow?context=linkedin/consumer/context#application-is-rejected
      if error in ('user_cancelled_login', 'user_cancelled_authorize'):
        logger.info(f"User declined: {request.values.get('error_description')}")
        return self.finish(None, state=request.values.get('state'))
      else:
        flask_util.error(f'{error} {desc}')

    # extract auth code and request access token
    auth_code = request.values['code']
    data = {
      'grant_type': 'authorization_code',
      'code': auth_code,
      'client_id': LINKEDIN_CLIENT_ID,
      'client_secret': LINKEDIN_CLIENT_SECRET,
      # redirect_uri here must be the same in the oauth code request!
      # (the value here doesn't actually matter since it's requested server side.)
      'redirect_uri': request.base_url,
    }

    resp = util.requests_post(ACCESS_TOKEN_URL, data=data)
    resp.raise_for_status()
    resp = json_loads(resp.text)

    logger.debug(f'Access token response: {resp}')
    if resp.get('serviceErrorCode'):
      flask_util.error(resp)

    access_token = resp['access_token']
    resp = LinkedInAuth(access_token_str=access_token).get(API_PROFILE_URL).json()
    logger.debug(f'Profile response: {resp}')
    auth = LinkedInAuth(id=resp['id'], access_token_str=access_token,
                        user_json=json_dumps(resp))
    auth.put()

    return self.finish(auth, state=request.values.get('state'))



"""Facebook OAuth drop-in.

https://developers.facebook.com/docs/facebook-login/manually-build-a-login-flow

TODO: implement client state param
TODO: unify this with instagram. see file docstring comment there.
"""
import logging
import urllib.error, urllib.parse

from flask import request
from google.cloud import ndb

from . import views, models
from .webutil import appengine_info, flask_util, util
from .webutil.util import json_dumps, json_loads

logger = logging.getLogger(__name__)

MIN_APP_SCOPED_ID = 100000000000000

API_BASE = 'https://graph.facebook.com/v4.0/'
if appengine_info.DEBUG:
  FACEBOOK_APP_ID = util.read('facebook_app_id_local')
  FACEBOOK_APP_SECRET = util.read('facebook_app_secret_local')
else:
  FACEBOOK_APP_ID = util.read('facebook_app_id')
  FACEBOOK_APP_SECRET = util.read('facebook_app_secret')
# facebook api url templates. can't (easily) use urllib.urlencode() because i
# want to keep the %(...)s placeholders as is and fill them in later in code.
# https://developers.facebook.com/docs/facebook-login/manually-build-a-login-flow#logindialog
GET_AUTH_CODE_URL = '&'.join((
    'https://www.facebook.com/v4.0/dialog/oauth?'
    # https://developers.facebook.com/docs/reference/login/
    'scope=%(scope)s',
    'client_id=%(client_id)s',
    # redirect_uri here must be the same in the access token request!
    'redirect_uri=%(redirect_uri)s',
    'state=%(state)s',
    'response_type=code',
))
# https://developers.facebook.com/docs/facebook-login/manually-build-a-login-flow#exchangecode
GET_ACCESS_TOKEN_URL = '&'.join((
    API_BASE + 'oauth/access_token?'
    'client_id=%(client_id)s',
    # redirect_uri here must be the same in the oauth request!
    # (the value here doesn't actually matter since it's requested server side.)
    'redirect_uri=%(redirect_uri)s',
    'client_secret=%(client_secret)s',
    'code=%(auth_code)s',
))
API_USER_URL = API_BASE + 'me?fields=id,email,name,picture'
API_PAGE_URL = API_BASE + 'me?fields=id,about,cover,description,emails,general_info,is_published,link,location,name,personal_info,phone,username,website'
API_PAGES_URL = API_BASE + 'me/accounts'


class FacebookAuth(models.BaseAuth):
  """An authenticated Facebook user or page.

  Provides methods that return information about this user (or page) and make
  OAuth-signed requests to Facebook's HTTP-based APIs. Stores OAuth credentials
  in the datastore. See models.BaseAuth for usage details.

  Facebook-specific details: implements urlopen() but not api(). The
  key name is the user's or page's Facebook ID.
  """
  type = ndb.StringProperty(choices=('user', 'page'))
  auth_code = ndb.StringProperty()
  access_token_str = ndb.StringProperty(required=True)
  # https://developers.facebook.com/docs/graph-api/reference/user#fields
  user_json = ndb.TextProperty(required=True)
  # https://developers.facebook.com/docs/graph-api/reference/user/accounts#fields
  pages_json = ndb.TextProperty()

  def site_name(self):
    return 'Facebook'

  def user_display_name(self):
    """Returns the user's or page's name.
    """
    return json_loads(self.user_json)['name']

  def access_token(self):
    """Returns the OAuth access token string.
    """
    return self.access_token_str

  def urlopen(self, url, **kwargs):
    """Wraps urlopen() and adds OAuth credentials to the request.
    """
    return models.BaseAuth.urlopen_access_token(url, self.access_token_str,
                                                **kwargs)

  def for_page(self, page_id):
    """Returns a new, unsaved FacebookAuth entity for a page in pages_json.

    The returned entity's properties will be populated with the page's data.
    access_token will be the page access token, user_json will be the page
    object, and pages_json will be a single-element list with the page.

    If page_id is not in pages_json, returns None.

    Args:
      page_id: string, Facebook page id
    """
    for page in json_loads(self.pages_json):
      id = page.get('id')
      if id == page_id:
        entity = FacebookAuth(id=id, type='page', pages_json=json_dumps([page]),
                              access_token_str=page.get('access_token'))
        entity.user_json = entity.urlopen(API_PAGE_URL).read()
        logger.debug(f'Page object: {entity.user_json}')
        return entity

    return None

  def is_authority_for(self, key):
    """Additionally check if the key represents a Page that this user has
    authority over.

    Args:
      auth_entity_key: ndb.Key

    Returns:
      boolean: true if key represents this user or one of the user's pages.
    """
    return super().is_authority_for(key) or any(
      key == self.for_page(page.get('id')).key
      for page in json_loads(self.pages_json))


class Start(views.Start):
  """Starts Facebook auth. Requests an auth code and expects a redirect back.
  """
  NAME = 'facebook'
  LABEL = 'Facebook'

  def redirect_url(self, state=None, app_id=None):
    if app_id is None:
      assert FACEBOOK_APP_ID and FACEBOOK_APP_SECRET, \
        "Please fill in the facebook_app_id and facebook_app_secret files in your app's root directory."
      app_id = FACEBOOK_APP_ID

    return GET_AUTH_CODE_URL % {
      'client_id': app_id,
      'scope': self.scope,
      # TODO: CSRF protection identifier.
      # http://developers.facebook.com/docs/authentication/
      'redirect_uri': urllib.parse.quote_plus(self.to_url()),
      'state': state,
    }


class Callback(views.Callback):
  """The auth callback. Fetches an access token, stores it, and redirects home."""
  def dispatch_request(self):
    if Callback.handle_error(self):
      return

    auth_code = request.values['code']
    url = GET_ACCESS_TOKEN_URL % {
      'auth_code': auth_code,
      'client_id': FACEBOOK_APP_ID,
      'client_secret': FACEBOOK_APP_SECRET,
      'redirect_uri': urllib.parse.quote_plus(request.base_url),
    }
    try:
      resp = json_loads(util.urlopen(url).read())
    except urllib.error.HTTPError as e:
      logger.error(e.read())
      raise

    logger.debug(f'Access token response: {resp}')
    access_token = resp['access_token']

    user = models.BaseAuth.urlopen_access_token(API_USER_URL, access_token).read()
    logger.debug(f'User info response: {user}')
    user_id = json_loads(user)['id']

    pages = json_dumps(json_loads(models.BaseAuth.urlopen_access_token(
      API_PAGES_URL, access_token).read()).get('data'))
    logger.debug(f'Pages response: {pages}')

    auth = FacebookAuth(id=user_id,
                        type='user',
                        user_json=user,
                        pages_json=pages,
                        auth_code=auth_code,
                        access_token_str=access_token)
    auth.put()
    return self.finish(auth, state=request.values.get('state'))

  @staticmethod
  def handle_error(handler):
    """Handles any error reported in the callback query parameters.

    Args:
      handler: Callback

    Returns:
      :class:`flask.Response` if there was an error, None otherwise.
    """
    error = request.values.get('error')
    error_reason = request.values.get('error_reason')

    if error or error_reason:
      error_description = urllib.parse.unquote_plus(
        request.values.get('error_description', ''))
      if error == 'access_denied' and error_reason == 'user_denied':
        logger.info(f'User declined: {error_description}')
        return handler.finish(None, state=request.values.get('state'))
      else:
        flask_util.error(' '.join((error, error_reason, error_description)))



"""Utilities for Flask. View classes, decorators, URL route converters, etc."""
import functools
import logging
import os
import re
import urllib.parse

import flask
from flask import abort, get_flashed_messages, make_response, redirect, render_template, request
from flask.views import View
from google.cloud import ndb
import werkzeug.exceptions
from werkzeug.exceptions import BadRequestKeyError, HTTPException
from werkzeug.routing import BaseConverter

from . import util

logger = logging.getLogger(__name__)

# Modern HTTP headers for CORS, CSP, other security, etc.
MODERN_HEADERS = {
  'Access-Control-Allow-Headers': '*',
  'Access-Control-Allow-Methods': '*',
  'Access-Control-Allow-Origin': '*',
  # see https://content-security-policy.com/
  'Content-Security-Policy':
    "script-src https: localhost:8080 my.dev.com:8080 'unsafe-inline'; "
    "frame-ancestors 'self'; ",
  # 16070400 seconds is 6 months
  'Strict-Transport-Security': 'max-age=16070400; preload',
  'X-Content-Type-Options': 'nosniff',
  # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Frame-Options
  'X-Frame-Options': 'SAMEORIGIN',
  'X-XSS-Protection': '1; mode=block',
}


# A few extra non-error HTTPExceptions
class Created(HTTPException):
    code = 201
    description = 'Created'

class Accepted(HTTPException):
    code = 202
    description = 'Accepted'

class NoContent(HTTPException):
    code = 204
    description = 'No Content'

class Redirect(HTTPException):
    def __init__(self, *args, location=None, **kwargs):
      assert location
      self.location = location
      super().__init__(**kwargs)

    def get_headers(self, *args, **kwargs):
      return {'Location': self.location}

class MovedPermanently(Redirect):
    code = 301
    description = 'Moved Permanently'

class Found(Redirect):
    code = 302
    description = 'Found'

class NotModified(HTTPException):
    code = 304
    description = 'Not Modified'

class PaymentRequired(HTTPException):
    code = 402
    description = 'Payment Required'

class ProxyAuthenticationRequired(HTTPException):
    code = 407
    description = 'Proxy Authentication Required'

class MisdirectedRequest(HTTPException):
    code = 421
    description = 'Misdirected Request'

class UpgradeRequired(HTTPException):
    code = 426
    description = 'Upgrade Required'

class PreconditionRequired(HTTPException):
    code = 428
    description = 'Precondition Required'

class ClientClosedRequest(HTTPException):
    code = 499
    description = 'Client Closed Request'

class VariantAlsoNegotiates(HTTPException):
    code = 506
    description = 'Variant Also Negotiates'

class InsufficientStorage(HTTPException):
    code = 507
    description = 'Insufficient Storage'

class LoopDetected(HTTPException):
    code = 508
    description = 'Loop Detected'

class NotExtended(HTTPException):
    code = 510
    description = 'Not Extended'

class NetworkAuthenticationRequired(HTTPException):
    code = 511
    description = 'Network Authentication Required'

class NetworkConnectTimeoutError(HTTPException):
    code = 599
    description = 'Network Connect Timeout Error'


for cls in (
    Created,
    Accepted,
    NoContent,
    MovedPermanently,
    Found,
    NotModified,
    PaymentRequired,
    ProxyAuthenticationRequired,
    MisdirectedRequest,
    UpgradeRequired,
    PreconditionRequired,
    ClientClosedRequest,
    VariantAlsoNegotiates,
    InsufficientStorage,
    LoopDetected,
    NotExtended,
    NetworkAuthenticationRequired,
    NetworkConnectTimeoutError,
):
  # https://github.com/pallets/flask/issues/1837#issuecomment-304996942
  werkzeug.exceptions.default_exceptions.setdefault(cls.code, cls)
  werkzeug.exceptions._aborter.mapping.setdefault(cls.code, cls)


class RegexConverter(BaseConverter):
  """Regexp URL route for Werkzeug/Flask.

  Based on https://github.com/rhyselsmore/flask-reggie.

  Usage:

    @app.route('/<regex("abc|def"):letters>')

  Install with:

    app = Flask(...)
    app.url_map.converters['regex'] = RegexConverter
  """
  def __init__(self, url_map, *items):
    super(RegexConverter, self).__init__(url_map)
    self.regex = items[0]


def get_required_param(name):
  """Returns the given request parameter.

  If it's not in a query parameter or POST field, the current HTTP request
  aborts with status 400.
  """
  try:
    val = request.values.get(name)
  except (UnicodeDecodeError, UnicodeEncodeError) as e:
    abort(400, f"Couldn't decode parameters as UTF-8: {e}")

  if not val:
    abort(400, f'Missing required parameter: {name}')

  return val


def ndb_context_middleware(app, client=None, **kwargs):
  """WSGI middleware to add an NDB context per request.

  Follows the WSGI standard. Details: http://www.python.org/dev/peps/pep-0333/

  Install with e.g.:

    ndb_client = ndb.Client()
    app = Flask('my-app')
    app.wsgi_app = flask_util.ndb_context_middleware(app.wsgi_app, ndb_client)

  Background: https://cloud.google.com/appengine/docs/standard/python3/migrating-to-cloud-ndb#using_a_runtime_context_with_wsgi_frameworks

  Args:
    client: :class:`google.cloud.ndb.Client`
    kwargs: passed through to :meth:`google.cloud.ndb.Client.context`
  """
  def wrapper(environ, start_response):
    if ndb.context.get_context(raise_context_error=False):
      # someone else (eg a unit test harness) has already created a context
      return app(environ, start_response)

    with client.context(**kwargs):
      return app(environ, start_response)

  return wrapper


def handle_exception(e):
  """Flask error handler that propagates HTTP exceptions into the response.

  Install with:
    app.register_error_handler(Exception, handle_exception)
  """
  if isinstance(e, BadRequestKeyError):
      if e.args:
          e._description = f'Missing required parameter: {e.args[0]}'
      else:
          e.show_exception = True

  logger.error(f'{e.__class__}: {e}')

  if isinstance(e, HTTPException):
    # raised by this app itself, pass it through. use body and headers from
    # response if available (but not status code).
    resp = e.get_response()
    if resp:
      resp.status_code = e.code
      return resp
    else:
      return str(e), e.code

  code, body = util.interpret_http_exception(e)
  if code:
    return ((f'Upstream server request failed: {e}' if code in ('502', '504')
             else f'HTTP Error {code}: {body}'),
            int(code),
            {'Content-Type': 'text/plain; charset=utf-8'})

  raise e


def error(msg, status=400, exc_info=False, **kwargs):
  """Logs and returns an HTTP error via :class:`werkzeug.exceptions.HTTPException`.

  Args:
    msg: str
    status: int
    exc_info: Python exception info three-tuple, eg from sys.exc_info()
    kwargs: passed through to :func:`flask.abort`
  """
  logger.info(f'Returning {status}: {msg} {kwargs}', exc_info=exc_info)
  abort(int(status), msg, **kwargs)


def flash(msg, **kwargs):
  """Wrapper for :func:`flask.flash` that also logs the message."""
  flask.flash(msg, **kwargs)
  logger.info(f'Flashed message: {msg}')


def default_modern_headers(resp):
  """Include modern HTTP headers by default, but let the response override them.

  Install with:
    app.after_request(default_modern_headers)
  """
  for name, value in MODERN_HEADERS.items():
    resp.headers.setdefault(name, value)

  return resp


def cached(cache, timeout, headers=(), http_5xx=False):
  """Thin flask-cache wrapper that supports timedelta and cache query param.

  If the `cache` URL query parameter is `false`, skips the cache. Also, does not
  store the response in the cache if it's an HTTP 5xx or if there are any
  flashed messages.

  Args:
    cache: :class:`flask_caching.Cache`
    timeout: :class:`datetime.timedelta`
    headers: sequence of str, optional headers to include in the cache key
    http_5xx: bool, optional, whether to cache HTTP 5xx (server error) responses
  """
  def response_filter(resp):
    """Return False if the response shouldn't be cached."""
    resp = make_response(resp)
    return (not get_flashed_messages() and 'Set-Cookie' not in resp.headers and
            (http_5xx or resp.status_code // 100 != 5))

  def unless():
    return bool(request.args.get('cache', '').lower() == 'false' or
                request.cookies)

  def decorator(f):
    # catch werkzeug HTTPExceptions, eg raised by abort(), and return them
    # instead of letting them propagate, so that flask-cache can cache them
    @functools.wraps(f)
    def httpexception_to_return(*args, **kwargs):
      try:
        return f(*args, **kwargs)
      except HTTPException as e:
        return e

    decorated = cache.cached(timeout.total_seconds(), query_string=True,
                             response_filter=response_filter, unless=unless
                             )(httpexception_to_return)

    # include specified headers in cache key:
    # https://flask-caching.readthedocs.io/en/latest/api.html#flask_caching.Cache.cached
    orig_cache_key = decorated.make_cache_key
    def make_cache_key(*args, **kwargs):
      header_vals = '  '.join(request.headers.get(h, '') for h in sorted(headers))
      k = f'{orig_cache_key(*args, **kwargs)}  {header_vals}'
      if request.method != 'GET':
        k = f'{request.method} {k}'
      return k

    decorated.make_cache_key = make_cache_key

    return decorated

  return decorator


def canonicalize_domain(from_domains, to_domain):
  """Returns a callable that redirects one or more domains to a canonical domain.

  Preserves scheme, path, and query.

  Install with eg:

    app = flask.Flask(...)
    app.before_request(canonicalize_domain(('old1.com', 'old2.org'), 'new.com'))

  Args:
    from_domains: str or sequence of str
    to_domain: str
  """
  if isinstance(from_domains, str):
    from_domains = [from_domains]

  def fn():
    parts = list(urllib.parse.urlparse(request.url))
    # not using request.host because it includes port
    if parts[1] in from_domains:  # netloc
      parts[1] = to_domain
      return redirect(urllib.parse.urlunparse(parts), code=301)

  return fn


class XrdOrJrd(View):
  """Renders and serves an XRD or JRD file.

  JRD is served if the request path ends in .jrd or .json, or the format query
  parameter is 'jrd' or 'json', or the request's Accept header includes 'jrd' or
  'json'.

  XRD is served if the request path ends in .xrd or .xml, or the format query
  parameter is 'xml' or 'xrd', or the request's Accept header includes 'xml' or
  'xrd'.

  Otherwise, defaults to DEFAULT_TYPE.

  Subclasses must override :meth:`template_prefix()` and
  :meth:`template_vars()`. URL route variables are passed through to
  :meth:`template_vars()` as keyword args.

  Class members:
    DEFAULT_TYPE: either JRD or XRD, which type to return by default if the
    request doesn't ask for one explicitly with the Accept header.
  """
  JRD = 'jrd'
  XRD = 'xrd'
  DEFAULT_TYPE = JRD  # either JRD or XRD

  def template_prefix(self):
    """Returns template filename, without extension."""
    raise NotImplementedError()

  def template_vars(self, **kwargs):
    """Returns a dict with template variables.

    URL route variables are passed through as kwargs.
    """
    raise NotImplementedError()

  def _type(self):
    """Returns XRD or JRD."""
    format = request.args.get('format', '').lower()
    ext = os.path.splitext(request.path)[1]

    if ext in ('.jrd', '.json') or format in ('jrd', 'json'):
      return self.JRD
    elif ext in ('.xrd', '.xml') or format in ('xrd', 'xml'):
      return self.XRD

    # We don't do full content negotiation (Accept Header parsing); we just
    # check whether jrd/json and xrd/xml are in the header, and if they both
    # are, which one comes first. :/
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Content_negotiation
    accept = request.headers.get('Accept', '').lower()
    jrd = re.search(r'jrd|json', accept)
    xrd = re.search(r'xrd|xml', accept)
    if jrd and (not xrd or jrd.start() < xrd.start()):
      return self.JRD
    elif xrd and (not jrd or xrd.start() < jrd.start()):
      return self.XRD

    assert self.DEFAULT_TYPE in (self.JRD, self.XRD)
    return self.DEFAULT_TYPE

  def dispatch_request(self, **kwargs):
    data = self.template_vars(**kwargs)
    assert isinstance(data, dict)

    # Content-Types are from https://tools.ietf.org/html/rfc7033#section-10.2
    if self._type() == self.JRD:
        return data, {'Content-Type': 'application/jrd+json'}

    template = f'{self.template_prefix()}.{self._type()}'
    return (render_template(template, **data),
            {'Content-Type': 'application/xrd+xml; charset=utf-8'})



"""IndieAuth drop-in.

https://indieauth.com/developers
"""
import logging
import urllib.parse

from flask import request
from google.cloud import ndb
import mf2util
import pkce
import requests

from . import models, views
from .webutil import flask_util, util
from .webutil.util import json_dumps, json_loads

logger = logging.getLogger(__name__)

INDIEAUTH_CLIENT_ID = util.read('indieauth_client_id')
INDIEAUTH_URL = 'https://indieauth.com/auth'


def discover_endpoint(rel, resp):
  """Fetch a URL and look for the `rel` Link header or HTML value.

  Args:
    rel: string, rel name to look for
    resp: :class:`requests.Response` to look in

  Return:
    string, the discovered `rel` value, or None if no endpoint was discovered
  """
  # check endpoint header first
  endpoint = resp.links.get(rel, {}).get('url')
  if endpoint:
    return endpoint

  # check the html content
  soup = util.parse_html(resp.text)
  link = soup.find('link', {'rel': rel})
  if link:
    return link.get('href')


def build_user_json(me):
  """Returns a JSON dict with h-card, rel-me links, and me value.

  Args:
    me: string, URL of the user, returned by
    resp: :class:`requests.Response` to use

  Return:
    dict, with 'me', the URL for this person; 'h-card', the representative h-card
      for this page; 'rel-me', a list of rel-me URLs found at this page
  """
  user_json = {'me': me}

  resp = util.requests_get(me)
  if resp.status_code // 100 != 2:
    logger.warning(f'could not fetch user url {me}, got response {resp.status_code}')
    return user_json

  mf2 = util.parse_mf2(resp, resp.url)
  user_json.update({
    'rel-me': mf2['rels'].get('me'),
    'h-card': mf2util.representative_hcard(mf2, me),
  })
  logger.debug(f'built user-json {user_json!r}')
  return util.trim_nulls(user_json)


class IndieAuth(models.BaseAuth):
  """An authenticated IndieAuth user.

  Provides methods that return information about this user. Stores credentials
  in the datastore. Key is the authed `me` URL value. See models.BaseAuth for
  usage details.
  """
  user_json = ndb.TextProperty(required=True)  # generally this has only 'me'
  access_token_str = ndb.StringProperty()
  refresh_token_str = ndb.StringProperty()

  def site_name(self):
    return 'IndieAuth'

  def user_display_name(self):
    """Returns the user's domain."""
    return self.key_id()

  def access_token(self):
    """Return the access token, N/A for IndieAuth"""
    return self.access_token_str


class Start(views.Start):
  """Starts the IndieAuth flow. Requires the 'me' parameter with the
  user URL that we want to authenticate.
  """
  NAME = 'indieauth'
  LABEL = 'IndieAuth'

  def redirect_url(self, state=None, me=None):
    assert INDIEAUTH_CLIENT_ID, (
      "Please fill in the indieauth_client_id in your app's root directory.")

    # TODO: unify with mastodon?
    if not me:
      me = request.values['me']
    parsed = urllib.parse.urlparse(me)
    if not parsed.scheme:
      me = 'http://' + me

    # fetch user URL
    redirect_uri = self.to_url()
    try:
      resp = util.requests_get(me)
    except (ValueError, requests.URLRequired, requests.TooManyRedirects) as e:
      flask_util.error(str(e))

    # discover endpoints
    if resp.ok:
      token_endpoint = discover_endpoint('token_endpoint', resp)
      auth_endpoint = (discover_endpoint('authorization_endpoint', resp) or
                       INDIEAUTH_URL)
    else:
      logger.warning(f'could not fetch user url {me}, got response {resp.status_code}')
      auth_endpoint = INDIEAUTH_URL
      token_endpoint = None

    # construct redirect URL
    if token_endpoint:
      code_verifier, code_challenge = pkce.generate_pkce_pair()
      return auth_endpoint + '?' + urllib.parse.urlencode({
        'me': me,
        'client_id': INDIEAUTH_CLIENT_ID,
        'redirect_uri': redirect_uri,
        'scope': 'profile',
        'code_challenge': code_challenge,
        'code_challenge_method': 'S256',
        'response_type': 'code',
        'state': util.encode_oauth_state({
          'code_verifier': code_verifier,
          'token_endpoint': token_endpoint,
          'me': me,
          'state': state,
          }),
        })
    else:
      return auth_endpoint + '?' + urllib.parse.urlencode({
        'me': me,
        'client_id': INDIEAUTH_CLIENT_ID,
        'redirect_uri': redirect_uri,
        'state': util.encode_oauth_state({
          'endpoint': auth_endpoint,
          'me': me,
          'state': state,
        }),
      })

  @classmethod
  def button_html(cls, *args, **kwargs):
    return super(cls, cls).button_html(
      *args,
      input_style='background-color: #EBEBEB; padding: 5px',
      form_extra='<input type="url" name="me" class="form-control" placeholder="Your web site" required style="width: 150px; height: 50px; display:inline;" />',
      **kwargs)


class Callback(views.Callback):
  """The callback view from the IndieAuth request. Performs an Authorization
  Code grant to verify the code."""
  def dispatch_request(self):
    code = request.values['code']
    state = util.decode_oauth_state(request.values['state'])

    token_endpoint = state.get('token_endpoint')
    me = state.get('me')
    if token_endpoint:
      # TODO: validate that the `iss` matches the value that is retrieved from the IndieAuth Server Metadata https://indieauth.spec.indieweb.org/#indieauth-server-metadata
      code_verifier = state.get('code_verifier') or ''
      state = state.get('state') or ''
      validate_resp = util.requests_post(token_endpoint, data={
        'grant_type': 'authorization_code',
        'client_id': INDIEAUTH_CLIENT_ID,
        'code': code,
        'redirect_uri': request.base_url,
        'code_verifier': code_verifier,
      })
    else:
      endpoint = state.get('endpoint')
      me = state.get('me')
      if not endpoint or not me:
        flask_util.error("invalid state parameter")

      state = state.get('state') or ''
      validate_resp = util.requests_post(endpoint, data={
        'me': me,
        'client_id': INDIEAUTH_CLIENT_ID,
        'code': code,
        'redirect_uri': request.base_url,
        'state': state,
      })

    if validate_resp.ok:
      data = util.sniff_json_or_form_encoded(validate_resp.text)
      if data.get('me'):
        verified = data.get('me')
        user_json = build_user_json(verified)
        indie_auth = IndieAuth(id=verified,
                               user_json=json_dumps(user_json),
                               access_token_str=data.get('access_token'),
                               refresh_token_str=data.get('refresh_token'),
                               )
        indie_auth.put()
        return self.finish(indie_auth, state=state)
      else:
        flask_util.error('Verification response missing required "me" field')
    else:
      flask_util.error(f'IndieAuth verification failed: {validate_resp.status_code} {validate_resp.text}')



"""Meetup.com drop-in.

API docs:
https://www.meetup.com/meetup_api/
"""
import logging
import urllib.parse

from flask import request
from google.cloud import ndb

from . import views
from .models import BaseAuth
from .webutil import appengine_info, flask_util, util
from .webutil.util import json_loads

logger = logging.getLogger(__name__)

if appengine_info.DEBUG:
    MEETUP_CLIENT_ID = util.read('meetup_client_id_local')
    MEETUP_CLIENT_SECRET = util.read('meetup_client_secret_local')
else:
    MEETUP_CLIENT_ID = util.read('meetup_client_id')
    MEETUP_CLIENT_SECRET = util.read('meetup_client_secret')

GET_AUTH_CODE_URL = '&'.join((
    'https://secure.meetup.com/oauth2/authorize?client_id=%(client_id)s',
    'response_type=code',
    'redirect_uri=%(redirect_uri)s',
    'scope=%(scope)s',
    'state=%(state)s',
))

GET_ACCESS_TOKEN_URL = 'https://secure.meetup.com/oauth2/access'
GET_USER_INFO_URL = 'https://api.meetup.com/members/self/'


def urlopen_bearer_token(url, access_token, data=None, **kwargs):
    """Wraps urlopen() and adds OAuth credentials to the request.
    """
    headers = {'Authorization': f'Bearer {access_token}'}
    try:
        return util.urlopen(urllib.request.Request(url, headers=headers, data=data), **kwargs)
    except BaseException as e:
        util.interpret_http_exception(e)
        raise


class MeetupAuth(BaseAuth):
    """An authenticated Meetup.com user.

    Provides methods that return information about this user and make
    OAuth-signed requests to Meetup's HTTP-based APIs. Stores OAuth credentials
    in the datastore. See models.BaseAuth for usage details.

    Implements urlopen() but not api().
    """
    access_token_str = ndb.StringProperty(required=True)
    user_json = ndb.TextProperty(required=True)

    def site_name(self):
        return 'Meetup.com'

    def user_display_name(self):
        """Returns the Meetup.com user id.
        """
        return json_loads(self.user_json)['name']

    def access_token(self):
        """Returns the OAuth access token string.
        """
        return self.access_token_str

    def urlopen(self, url, **kwargs):
        return urlopen_bearer_token(url, self.access_token_str, kwargs)


class MeetupCsrf(ndb.Model):
    """Stores a CSRF token for the Meetup.com OAuth2 flow."""
    token = ndb.StringProperty(required=False)
    state = ndb.TextProperty(required=False)


class Start(views.Start):
    """Starts Meetup.com auth. Requests an auth code and expects a redirect back.
    """

    NAME = 'meetup'
    LABEL = 'Meetup.com'
    DEFAULT_SCOPE = ''

    def redirect_url(self, state=None):

        assert MEETUP_CLIENT_ID and MEETUP_CLIENT_SECRET, \
            "Please fill in the meetup_client_id and meetup_client_secret files in your app's root directory."

        csrf_key = MeetupCsrf(state=state).put()

        return GET_AUTH_CODE_URL % {
            'client_id': MEETUP_CLIENT_ID,
            'redirect_uri': urllib.parse.quote_plus(self.to_url()),
            'scope': self.scope,
            'state': f'{urllib.parse.quote_plus(state or "")}|{csrf_key.id()}',
        }

    @classmethod
    def button_html(cls, *args, **kwargs):
        return super(cls, cls).button_html(
            *args, input_style='background-color: #EEEEEE; padding: 10px', **kwargs)


class Callback(views.Callback):
    """The auth callback. Fetches an access token, stores it, and redirects home.
    """
    def dispatch_request(self):
        # handle errors
        error = request.values.get('error')
        if error:
            if error == 'access_denied':
                logger.info('User declined')
                return self.finish(None, state=request.values.get('state'))
            else:
                flask_util.error(f"Error: {error}: {request.values.get('error_description')}")

        state = request.values['state']
        # lookup the CSRF token
        try:
            csrf_id = int(urllib.parse.unquote_plus(state).split('|')[-1])
        except (ValueError, TypeError):
            flask_util.error(f'Invalid state value {state!r}')

        csrf = MeetupCsrf.get_by_id(csrf_id)
        if not csrf:
            flask_util.error(f'No CSRF token for id {csrf_id}')

        # extract auth code and request access token
        auth_code = request.values['code']
        data = {
            'code': auth_code,
            'client_id': MEETUP_CLIENT_ID,
            'client_secret': MEETUP_CLIENT_SECRET,
            'grant_type': 'authorization_code',
            # redirect_uri here must be the same in the oauth code request!
            # (the value here doesn't actually matter since it's requested server side.)
            'redirect_uri': request.base_url,
        }
        # TODO: handle refresh tokens
        resp = util.requests_post(GET_ACCESS_TOKEN_URL, data=data)
        resp.raise_for_status()

        logger.debug(f'Access token response: {resp}')

        try:
            data = json_loads(resp.text)
        except (ValueError, TypeError):
            logger.error(f'Bad response:\n{resp}', exc_info=True)
            flask_util.error('Bad Disqus response to access token request')

        error = data.get('error')
        if error:
            msg = f"Error: {error[0]}: {data.get('error_description')}"
            flask_util.error(msg)

        access_token = data['access_token']

        user_json = urlopen_bearer_token(GET_USER_INFO_URL, access_token).read()
        user = json_loads(user_json)
        logger.debug(f'User info response: {user}')
        user_id = str(user['id'])

        logger.info(f'Storing new Meetup account for ID: {user_id}')
        auth = MeetupAuth(id=user_id, access_token_str=access_token, user_json=user_json)
        auth.put()
        return self.finish(auth, state=csrf.state)



"""Disqus OAuth drop-in.

Disqus API docs: https://disqus.com/api/docs/

This drop-in is even more similar to Instagram than Instagram is to
Facebook. Differences:

- urlopen must pass the api_key with each request (in addition to the
  access_token)
- Response to access_token does not give much information about the user,
  so we additionally fetch /user/details before saving
- Deny appears to be broken on Disqus's side (clicking "No Thanks" has
  no effect), so we ignore that possibility for now.

TODO unify Disqus, Facebook, and Instagram
"""
import logging
import urllib.parse

from flask import request
from google.cloud import ndb

from . import models, views
from .webutil import flask_util, util
from .webutil.util import json_dumps, json_loads

logger = logging.getLogger(__name__)

DISQUS_CLIENT_ID = util.read('disqus_client_id')
DISQUS_CLIENT_SECRET = util.read('disqus_client_secret')
GET_AUTH_CODE_URL = (
    'https://disqus.com/api/oauth/2.0/authorize/?' +
    '&'.join((
        'client_id=%(client_id)s',
        'scope=%(scope)s',
        'response_type=code',
        'redirect_uri=%(redirect_uri)s',
    )))
GET_ACCESS_TOKEN_URL = 'https://disqus.com/api/oauth/2.0/access_token/'
USER_DETAILS_URL = 'https://disqus.com/api/3.0/users/details.json?user=%d'


class DisqusAuth(models.BaseAuth):
  """An authenticated Disqus user.

  Provides methods that return information about this user (or page)
  and make OAuth-signed requests to Instagram's HTTP-based
  APIs. Stores OAuth credentials in the datastore. See models.BaseAuth
  for usage details.

  Disqus-specific details: implements urlopen() but not api().
  The key name is the Disqus user id.
  """
  auth_code = ndb.StringProperty(required=True)
  access_token_str = ndb.StringProperty(required=True)
  user_json = ndb.TextProperty(required=True)

  def site_name(self):
    return 'Disqus'

  def user_display_name(self):
    """Returns the user's name.
    """
    return json_loads(self.user_json)['name']

  def access_token(self):
    """Returns the OAuth access token string.
    """
    return self.access_token_str

  def urlopen(self, url, **kwargs):
    """Wraps urlopen() and adds OAuth credentials to the request.
    """
    # TODO does work for POST requests? key is always passed as a
    # query param, regardless of method.
    return models.BaseAuth.urlopen_access_token(url, self.access_token_str,
                                                DISQUS_CLIENT_ID, **kwargs)


class Start(views.Start):
  """Starts Disqus auth. Requests an auth code and expects a redirect back.
  """
  NAME = 'disqus'
  LABEL = 'Disqus'

  # Disqus scopes are comma separated: read, write, admin, email
  # https://disqus.com/api/docs/requests/#data-availability
  DEFAULT_SCOPE = 'read'

  def redirect_url(self, state=None):
      assert DISQUS_CLIENT_ID and DISQUS_CLIENT_SECRET, \
          "Please fill in the disqus_client_id and disqus_client_secret files in your app's root directory."
      return GET_AUTH_CODE_URL % {
        'client_id': DISQUS_CLIENT_ID,
        'scope': self.scope,
        'redirect_uri': urllib.parse.quote_plus(self.to_url(state=state)),
      }


class Callback(views.Callback):
  """The auth callback. Fetches an access token, stores it, and redirects home.
  """
  def dispatch_request(self):
    if self.handle_error():
      return

    # https://disqus.com/api/docs/auth/
    auth_code = request.values['code']
    data = {
        'grant_type': 'authorization_code',
        'client_id': DISQUS_CLIENT_ID,
        'client_secret': DISQUS_CLIENT_SECRET,
        'redirect_uri': self.request_url_with_state(),
        'code': auth_code,
    }

    resp = util.requests_post(GET_ACCESS_TOKEN_URL, data=data)
    resp.raise_for_status()
    try:
      data = json_loads(resp.text)
    except (ValueError, TypeError):
      logger.error(f'Bad response:\n{resp}', exc_info=True)
      flask_util.error('Bad Disqus response to access token request')

    access_token = data['access_token']
    user_id = data['user_id']
    # TODO is a username key preferred?
    # username = data['username']

    auth = DisqusAuth(id=str(user_id),
                      auth_code=auth_code,
                      access_token_str=access_token)

    resp = auth.urlopen(USER_DETAILS_URL % user_id).read()
    try:
      user_data = json_loads(resp)['response']
    except (ValueError, TypeError):
      logger.error(f'Bad response:\n{resp}', exc_info=True)
      flask_util.error('Bad Disqus response to user details request')

    auth.user_json = json_dumps(user_data)
    logger.info(f'created disqus auth {auth}')
    auth.put()
    return self.finish(auth, state=request.values.get('state'))

  def handle_error(handler):
    """Handles any error reported in the callback query parameters.

    Args:
      handler: Callback

    Returns:
      True if there was an error, False otherwise.
    """
    error = request.values.get('error')
    if error:
      if error == 'access_denied':
        logger.info('User declined')
        handler.finish(None, state=request.values.get('state'))
        return True
      else:
        flask_util.error(error)

    return False



"""Twitter OAuth 2 drop-in.

https://developer.twitter.com/en/docs/authentication/oauth-2-0/user-access-token
https://developer.twitter.com/en/docs/authentication/oauth-2-0/authorization-code
https://developer.twitter.com/en/docs/authentication/api-reference/token
"""
import logging
import secrets
import time

from flask import request
from google.cloud import ndb
from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session
from urllib.parse import quote_plus, unquote, urlencode, urljoin, urlparse, urlunparse

from . import models, views
from .webutil import flask_util, util
from .webutil.util import json_dumps, json_loads

logger = logging.getLogger(__name__)

TWITTER_CLIENT_ID = util.read('twitter_app_key')
TWITTER_CLIENT_SECRET = util.read('twitter_app_secret')

AUTH_CODE_URL = 'https://twitter.com/i/oauth2/authorize'
ACCESS_TOKEN_URL = 'https://api.twitter.com/2/oauth2/token'
API_ACCOUNT_URL = 'https://api.twitter.com/2/users/me'

# https://developer.twitter.com/en/docs/authentication/oauth-2-0/authorization-code
ALL_SCOPES = (
  'block.read',
  'block.write',
  'bookmark.read',
  'bookmark.write',
  'follows.read',
  'follows.write',
  'like.read',
  'like.write',
  'list.read',
  'list.write',
  'mute.read',
  'mute.write',
  'offline.access',
  'space.read',
  'tweet.read',
  'tweet.write',
  'users.read',
)


class TwitterOAuth2(models.BaseAuth):
  """An OAuth2-authenticated Twitter user.

  Provides methods that return information about this user and store OAuth 2 tokens
  in the datastore. See models.BaseAuth for usage details.

  The datastore entity key name is the Twitter username.
  """
  # Fields: token_type, access_token, scope, expires_at, expires_in
  token_json = ndb.TextProperty(required=True)
  user_json = ndb.TextProperty(required=True)

  def site_name(self):
    return 'Twitter'

  def user_display_name(self):
    """Returns the username."""
    return self.key_id()

  def access_token(self):
    """Returns the OAuth access token JSON."""
    return json_loads(self.token_json)['access_token']

  def session(self):
    """Returns a :class:`requests_oauthlib.OAuth2Session`."""
    token = json_loads(self.token_json)

    if token.get('refresh_token') and token.get('expires_at'):
      def update_token(token):
        logging.info(f'Storing new access token {token}')
        self.token = token
        self.put()

      kwargs = {
        'auto_refresh_url': ACCESS_TOKEN_URL,
        'auto_refresh_kwargs': {'client_id': TWITTER_CLIENT_ID},
        'token_updater': update_token,
      }

    session = OAuth2Session(TWITTER_CLIENT_ID, token=token, **kwargs)
    session.auth = HTTPBasicAuth(TWITTER_CLIENT_ID, TWITTER_CLIENT_SECRET)
    return session


class Start(views.Start):
  """Starts three-legged OAuth with Twitter.

  Redirects to Twitter's auth prompt for user approval.
  """
  NAME = 'twitter'
  LABEL = 'Twitter'
  SCOPE_SEPARATOR = ' '
  DEFAULT_SCOPE = 'tweet.read users.read'

  def redirect_url(self, state=None):
    assert TWITTER_CLIENT_ID and TWITTER_CLIENT_SECRET, \
      "Please fill in the twitter_app_key and twitter_app_secret files in your app's root directory."

    if not state:
      state = secrets.token_urlsafe(32)
      logging.debug(f'No state provided; generated default random state {state}')

    # generate and store PKCE code
    verifier = secrets.token_urlsafe(64)
    key = models.PkceCode(id=state, challenge=verifier, verifier=verifier).put()
    logging.info(f'Storing PKCE code verifier {verifier}: {key}')

    # redirect to Twitter auth URL
    session = OAuth2Session(TWITTER_CLIENT_ID, scope=self.scope,
                            redirect_uri=self.to_url())
    auth_url, state = session.authorization_url(
      AUTH_CODE_URL, state=state,
      code_challenge=verifier,
      code_challenge_method='plain')
    logger.info(f'Redirecting to {auth_url}')
    return auth_url


class Callback(views.Callback):
  """The OAuth callback. Fetches an access token and redirects to the front page.
  """
  def dispatch_request(self):
    state = request.values.get('state')
    error = request.values.get('error')
    desc = request.values.get('error_description')
    if error:
      msg = f'Error: {error}: {desc}'
      logger.info(msg)
      if error == 'access_denied':
        return self.finish(None, state=state)
      else:
        flask_util.error(msg)

    # look up PKCE code verifier
    code = models.PkceCode.get_by_id(state)
    if not code:
      flask_util.error(f'state not found: {state}')
    logging.info(f'Loaded PKCE code {code}')

    session = OAuth2Session(TWITTER_CLIENT_ID, redirect_uri=request.base_url)
    session.fetch_token(ACCESS_TOKEN_URL, code=request.values['code'],
                        client_secret=TWITTER_CLIENT_SECRET,
                        authorization_response=request.url,
                        code_verifier=code.verifier)
    logging.info(f'Got access token {session.token}')

    # Fetch user info
    resp = util.requests_get(API_ACCOUNT_URL, session=session)
    resp.raise_for_status()
    user_json = resp.json()
    logging.info(f'{user_json}')
    username = user_json['data']['username']

    auth = TwitterOAuth2(id=username, token_json=json_dumps(session.token),
                         user_json=json_dumps(user_json))
    auth.put()

    return self.finish(auth, state=request.values.get('state'))



"""HTTP request and server information derived from the environment.

App Engine:
* https://cloud.google.com/appengine/docs/standard/python3/runtime#environment_variables
* https://cloud.google.com/appengine/docs/flexible/python/runtime#environment_variables

WSGI: https://www.python.org/dev/peps/pep-3333/

To run against the prod datastore, make a service account, download its JSON
credentials file, then run eg:

env GOOGLE_APPLICATION_CREDENTIALS=service_account_creds.json FLASK_ENV=development flask run -p 8080
"""
import os, sys

project = os.getenv('GOOGLE_CLOUD_PROJECT') or os.getenv('GAE_APPLICATION') or ''
APP_ID = project.split('~')[-1]

creds = os.getenv('GOOGLE_APPLICATION_CREDENTIALS')
if creds and not creds.endswith('fake_user_account.json'):
  DEBUG = False
  LOCAL_SERVER = True
else:
  DEBUG = os.environ.get('GAE_ENV') in (None, 'localdev')  # 'standard' in production
  LOCAL_SERVER = (os.environ.get('GAE_ENV') != 'standard'  # App Engine Standard
                  and not os.environ.get('GAE_INSTANCE'))  # App Engine Flex



from atproto.exceptions import UnauthorizedError
from . import views, models
from flask import request
from google.cloud import ndb
from .webutil.util import json_dumps
import atproto

class Start(views.Start):
  """
  Starts Bluesky auth - used just to provide the button.
  """
  NAME = 'bluesky'
  LABEL = 'Bluesky'

class BlueskyAuth(models.BaseAuth):
  """
  An authenticated Bluesky user.
  """

  did = ndb.StringProperty(required=True)
  password = ndb.StringProperty(required=True)
  user_json = ndb.TextProperty(required=True)

  def site_name(self):
    return 'Bluesky'

  def access_token(self):
    """
    Returns did and password as a tuple in place of an OAuth key id/secret pair.
    """
    return (self.did, self.password)

  def _api(self):
    """
    Returns an atproto.Client.
    """
    (client, _) = BlueskyAuth._api_from_password(self.did, self.password)
    return client

  @staticmethod
  def _api_from_password(handle, password):
    client = atproto.Client()
    profile = client.login(handle, password)
    return (client, profile)


class Callback(views.Callback):
  """
  OAuth callback stub.
  """
  def dispatch_request(self):
    username = request.values['username']
    password = request.values['password']

    # get the did (portable user ID)
    try:
      # TODO migrate this to lexrpc
      (_, profile) = BlueskyAuth._api_from_password(username, password)
    except UnauthorizedError:
      return self.finish(None)
    user_json = json_dumps({
      '$type': profile._type,
      'did': profile.did,
      'handle': profile.handle,
      'avatar': profile.avatar,
      'banner': profile.banner,
      'displayName': profile.displayName,
      'description': profile.description,
    })
    auth = BlueskyAuth(
      id=profile.did,
      did=profile.did,
      password=password,
      user_json=user_json,
    )
    auth.put()
    return self.finish(auth)



"""App Engine datastore model base classes, properties, and utilites.
"""
from google.cloud import ndb

from oauth_dropins.webutil.util import json_dumps, json_loads


class StringIdModel(ndb.Model):
  """An ndb model class that requires a string id."""
  def put(self, *args, **kwargs):
    """Raises AssertionError if string id is not provided."""
    assert self.key and self.key.string_id(), 'string id required but not provided'
    return super(StringIdModel, self).put(*args, **kwargs)


class JsonProperty(ndb.TextProperty):
    """Fork of ndb's that subclasses TextProperty instead of BlobProperty.

    This makes values show up as normal, human-readable, serialized JSON in the
    web console.
    https://github.com/googleapis/python-ndb/issues/874#issuecomment-1442753255

    Duplicated in arroba:
    https://github.com/snarfed/arroba/blob/main/arroba/ndb_storage.py
    """
    def _validate(self, value):
        if not isinstance(value, dict):
            raise TypeError('JSON property must be a dict')

    def _to_base_type(self, value):
        as_str = json_dumps(value, separators=(',', ':'), ensure_ascii=True)
        return as_str.encode('ascii')

    def _from_base_type(self, value):
        if not isinstance(value, str):
            value = value.decode('ascii')
        return json_loads(value)


class ComputedJsonProperty(JsonProperty, ndb.ComputedProperty):
    """Custom ComputedProperty for JSON values that stores them as strings.

    ...instead of like StructuredProperty, with "entity" type, which bloats them
    unnecessarily in the datastore.
    """
    def __init__(self, *args, **kwargs):
        kwargs['indexed'] = False
        super().__init__(*args, **kwargs)



"""Twitter OAuth drop-in.

TODO: port to
http://code.google.com/p/oauth/source/browse/#svn%2Fcode%2Fpython . tweepy is
just a wrapper around that anyway.
"""
import logging

from flask import request
from google.cloud import ndb
import tweepy

from . import models, twitter_auth, views
from .webutil import flask_util, util
from .webutil.util import json_dumps, json_loads

logger = logging.getLogger(__name__)

API_ACCOUNT_URL = 'https://api.twitter.com/1.1/account/verify_credentials.json'


class TwitterAuth(models.BaseAuth):
  """An authenticated Twitter user.

  Provides methods that return information about this user and make OAuth-signed
  requests to the Twitter v1.1 API. Stores OAuth credentials in the datastore.
  See models.BaseAuth for usage details.

  Twitter-specific details: implements api(), get(), and post(). api() returns a
  tweepy.API; get() and post() wrap the corresponding requests methods. The
  datastore entity key name is the Twitter username.
  """
  # access token
  token_key = ndb.StringProperty(required=True)
  token_secret = ndb.StringProperty(required=True)
  user_json = ndb.TextProperty(required=True)

  def site_name(self):
    return 'Twitter'

  def user_display_name(self):
    """Returns the username.
    """
    return self.key_id()

  def access_token(self):
    """Returns the OAuth access token as a (string key, string secret) tuple.
    """
    return (self.token_key, self.token_secret)

  def urlopen(self, url, **kwargs):
    """Wraps urllib.request.urlopen() and adds an OAuth signature.
    """
    return twitter_auth.signed_urlopen(url, self.token_key, self.token_secret,
                                       **kwargs)

  def get(self, *args, **kwargs):
    """Wraps requests.get() and adds an OAuth signature.
    """
    oauth1 = twitter_auth.auth(self.token_key, self.token_secret)
    resp = util.requests_get(*args, auth=oauth1, **kwargs)
    try:
      resp.raise_for_status()
    except BaseException as e:
      util.interpret_http_exception(e)
      raise
    return resp

  def post(self, *args, **kwargs):
    """Wraps requests.post() and adds an OAuth signature.
    """
    oauth1 = twitter_auth.auth(self.token_key, self.token_secret)
    resp = util.requests_post(*args, auth=oauth1, **kwargs)
    try:
      resp.raise_for_status()
    except BaseException as e:
      util.interpret_http_exception(e)
      raise
    return resp

  def api(self):
    """Returns a tweepy.API.
    """
    return tweepy.API(twitter_auth.tweepy_auth(self.token_key, self.token_secret))


class Start(views.Start):
  """Starts three-legged OAuth with Twitter.

  Fetches an OAuth request token, then redirects to Twitter's auth page to
  request an access token.

  Attributes:
    access_type: optional, 'read' or 'write'. Passed through to Twitter as
      x_auth_access_type. If the twitter app has read/write or read/write/dm
      permissions, this lets you request a read-only token. Details:
      https://dev.twitter.com/docs/api/1/post/oauth/request_token
  """
  NAME = 'twitter'
  LABEL = 'Twitter'

  def __init__(self, to_path, scopes=None, access_type=None):
    super().__init__(to_path, scopes=scopes)
    assert access_type in (None, 'read', 'write'), \
        f'access_type must be "read" or "write"; got {access_type!r}'
    self.access_type = access_type

  def redirect_url(self, state=None):
    assert twitter_auth.TWITTER_APP_KEY and twitter_auth.TWITTER_APP_SECRET, \
      "Please fill in the twitter_app_key and twitter_app_secret files in your app's root directory."
    auth = tweepy.OAuth1UserHandler(twitter_auth.TWITTER_APP_KEY,
                                    twitter_auth.TWITTER_APP_SECRET,
                                    callback=self.to_url(state=state))

    # signin_with_twitter=True returns /authenticate instead of /authorize so
    # that Twitter doesn't prompt the user for approval if they've already
    # approved. Background: https://dev.twitter.com/discussions/1253
    #
    # Requires "Allow this application to be used to Sign in with Twitter"
    # to be checked in the app's settings on https://apps.twitter.com/
    #
    # Also, there's a Twitter API bug that makes /authenticate and
    # x_auth_access_type not play nice with each other. Work around that by only
    # using /authenticate if access_type isn't set.
    # https://dev.twitter.com/discussions/21281
    auth_url = auth.get_authorization_url(
      signin_with_twitter=not self.access_type, access_type=self.access_type)

    # store the request token for later use in the callback view
    models.OAuthRequestToken(id=auth.request_token['oauth_token'],
                             token_secret=auth.request_token['oauth_token_secret']
                             ).put()
    logger.info(f'Generated request token, redirecting to Twitter: {auth_url}')
    return auth_url


class Callback(views.Callback):
  """The OAuth callback. Fetches an access token and redirects to the front page.
  """
  def dispatch_request(self):
    # https://dev.twitter.com/docs/application-permission-model
    if request.values.get('denied'):
      return self.finish(None, state=request.values.get('state'))
    oauth_token = request.values.get('oauth_token', None)
    oauth_verifier = request.values.get('oauth_verifier', None)
    if oauth_token is None:
      flask_util.error('Missing required query parameter oauth_token.')

    # Lookup the request token
    request_token = models.OAuthRequestToken.get_by_id(oauth_token)
    if request_token is None:
      flask_util.error(f'Invalid oauth_token: {oauth_token}')

    # Rebuild the auth view
    auth = tweepy.OAuth1UserHandler(twitter_auth.TWITTER_APP_KEY,
                                    twitter_auth.TWITTER_APP_SECRET)
    auth.request_token = {'oauth_token': request_token.key.string_id(),
                          'oauth_token_secret': request_token.token_secret}

    # Fetch the access token
    access_token_key, access_token_secret = auth.get_access_token(oauth_verifier)
    user_json = twitter_auth.signed_urlopen(API_ACCOUNT_URL,
                                            access_token_key,
                                            access_token_secret).read()
    username = json_loads(user_json)['screen_name']

    auth = TwitterAuth(id=username,
                       token_key=access_token_key,
                       token_secret=access_token_secret,
                       user_json=user_json)
    auth.put()
    return self.finish(auth, state=request.values.get('state'))



"""Unit test utilities."""
from datetime import datetime, timezone
import difflib
import email
import io
import os
import pprint
import re
import traceback
import urllib.error, urllib.parse, urllib.request
import warnings

from bs4 import (
    GuessedAtParserWarning,
    MarkupResemblesLocatorWarning,
    XMLParsedAsHTMLWarning,
)
from mox3 import mox
import requests

from . import appengine_info
from . import util
from .util import json_dumps, json_loads, HTTP_TIMEOUT

RE_TYPE = (re.Pattern if hasattr(re, 'Pattern')  # python >=3.7
           else re._pattern_type)                # python <3.7

NOW = datetime(2022, 1, 2, 3, 4, 5, tzinfo=timezone.utc)

# don't truncate assertion error diffs
import unittest.util
unittest.util._MAX_LENGTH = 999999


def requests_response(body='', url=None, status=200, content_type=None,
                      redirected_url=None, headers=None, allow_redirects=None,
                      encoding=None):
    """
    Args:
      redirected_url: string URL or sequence of string URLs for multiple redirects
    """
    resp = requests.Response()

    if isinstance(body, (dict, list)):
      body = json_dumps(body, indent=2)
      if content_type is None:
        content_type = 'application/json'

    resp._text = body
    resp._content = body.encode('utf-8') if isinstance(body, str) else body
    resp.raw = io.BytesIO(resp._content)  # needed for close()
    resp.encoding = encoding if encoding is not None else 'utf-8'

    resp.url = url
    if redirected_url is not None:
      if allow_redirects is False:
        resp.headers['location'] = redirected_url
      else:
        if isinstance(redirected_url, str):
          redirected_url = [redirected_url]
        assert isinstance(redirected_url, (list, tuple))
        resp.url = redirected_url[-1]
        for u in [url] + redirected_url[:-1]:
          resp.history.append(requests.Response())
          resp.history[-1].url = u

    resp.status_code = status
    if content_type is None:
      content_type = 'text/html'
    elif content_type == 'None':
      content_type = None
    resp.headers['content-type'] = content_type
    if headers is not None:
      resp.headers.update(headers)

    return resp


def enable_flask_caching(app, cache):
  """Test case decorator that enables a flask_caching cache.

  Usage:

    from app import app, cache

    class FooTest(TestCase):
      @enable_flask_caching(app, cache)
      def test_foo(self):
        ..

  Args:
    app: :class:`flask.Flask` app
    cache: :class:`flask_caching.Cache`
  """
  def decorator(method):
    def wrapper(self, *args, **kwargs):
      orig_cache_type = app.config.get('CACHE_TYPE')
      try:
        app.config['CACHE_TYPE'] = 'SimpleCache'
        cache.init_app(app)
        method(self, *args, **kwargs)
      finally:
        app.config['CACHE_TYPE'] = orig_cache_type
        cache.init_app(app)

    return wrapper

  return decorator


class UrlopenResult(object):
  """A fake :func:`urllib.request.urlopen()` or :func:`urlfetch.fetch()` result object.
  """
  def __init__(self, status_code, content, url=None, headers={}):
    self.status_code = status_code
    self.content = io.StringIO(str(content))
    self.url = url
    self.headers = headers

  def read(self, length=-1):
    return self.content.read(length)

  def getcode(self):
    return self.status_code

  def geturl(self):
    return self.url

  def info(self):
    return email.message_from_string(
        '\n'.join(f'{key}: {val}' for key, val in self.headers.items()))


class Asserts(object):
  """Test case mixin class with extra assert helpers."""

  def assert_entities_equal(self, a, b, ignore=frozenset(), keys_only=False,
                            in_order=False):
    """Asserts that a and b are equivalent entities or lists of entities.

    ...specifically, that they have the same property values, and if they both
    have populated keys, that their keys are equal too.

    Args:
      a: :class:`ndb.Model` instances or lists of instances
      b: same
      ignore: sequence of strings, property names not to compare
      keys_only: boolean, if True only compare keys
      in_order: boolean. If False, all entities must have keys.
    """
    # all the __class__.__name__ hacks below are so we avoid importing ndb
    # from the app engine SDK, since this file needs to support Python 3.
    if not (isinstance(a, (list, tuple) or a.__class__.__name__ == 'Query')):
      a = [a]
    if not (isinstance(b, (list, tuple) or b.__class__.__name__ == 'Query')):
      b = [b]

    if not in_order:
      a = list(sorted(a, key=lambda e: e.key))
      b = list(sorted(b, key=lambda e: e.key))

    self.assertEqual(len(a), len(b),
                     f'Different lengths:\n expected {a}\n actual {b}')

    for x, y in zip(a, b):
      x_key = None
      try:
        x_key = x.key.flat()
        self.assertEqual(x_key, y.key.flat())
      except Exception as err:
        if err.__class__.__name__ in ('BadKeyError', 'NotSavedError',
                                      'AttributeError'):
          if keys_only:
            raise
        else:
          raise

      def props(e):
        all = e.to_dict()
        return {k: v for k, v in list(all.items()) if k not in ignore}

      if not keys_only:
        self.assert_equals(props(x), props(y), x_key)

  def entity_keys(self, entities):
    """Returns a list of keys for a list of entities.
    """
    return [e.key() for e in entities]

  def assert_equals(self, expected, actual, msg=None, in_order=False, ignore=()):
    """Pinpoints individual element differences in lists and dicts.

    If in_order is False, ignores order in lists and tuples.
    """
    try:
      self._assert_equals(expected, actual, in_order=in_order, ignore=ignore)
    except AssertionError as e:
      if not isinstance(expected, str):
        expected = pprint.pformat(expected)
      if not isinstance(actual, str):
        actual = pprint.pformat(actual)
      raise AssertionError(f"""{msg}: {''.join(e.args)}
Expected value:
{expected}
Actual value:
{actual}""") from None

  def _assert_equals(self, expected, actual, in_order=False, ignore=()):
    """Recursive helper for assert_equals().
    """
    key = None

    try:
      if isinstance(expected, RE_TYPE):
        if not re.match(expected, actual):
          self.fail(f"{expected!r} doesn't match {actual}")
      elif isinstance(expected, dict) and isinstance(actual, dict):
        for key in set(expected.keys()) | set(actual.keys()):
          if key not in ignore:
            self._assert_equals(expected.get(key), actual.get(key),
                                in_order=in_order, ignore=ignore)
      elif (isinstance(expected, (list, tuple, set)) and
            isinstance(actual, (list, tuple, set))):
        if not in_order:
          # use custom key because Python 3 dicts are not comparable :/
          def hash_or_json(x):
            try:
              return str(hash(x))
            except TypeError:
              return json_dumps(x, sort_keys=True)
          expected = sorted(expected, key=hash_or_json)
          actual = sorted(actual, key=hash_or_json)

        self.assertEqual(len(expected), len(actual),
                         f'Different lengths:\n expected {len(expected)}\n actual {len(actual)}')
        for key, (e, a) in enumerate(zip(expected, actual)):
          self._assert_equals(e, a, in_order=in_order, ignore=ignore)
      elif (isinstance(expected, str) and isinstance(actual, str) and
            '\n' in expected):
        self.assert_multiline_equals(expected, actual)
      else:
        self.assertEqual(expected, actual)

    except AssertionError as e:
      # fill in where this failure came from. this recursively builds,
      # backwards, all the way up to the root.
      args = (f'[{key}] ' if key is not None else '') + ''.join(e.args)
      raise AssertionError(args) from None

  def assert_multiline_equals(self, expected, actual, ignore_blanks=False):
    """Compares two multi-line strings and reports a diff style output.

    Ignores leading and trailing whitespace on each line, and squeezes repeated
    blank lines down to just one.

    Args:
      ignore_blanks: boolean, whether to ignore blank lines altogether
    """
    exp = self._normalize_lines(expected, ignore_blanks=ignore_blanks)
    act = self._normalize_lines(actual, ignore_blanks=ignore_blanks)
    if exp != act:
      self.fail(''.join(difflib.Differ().compare(exp, act)))

  def assert_multiline_in(self, expected, actual, ignore_blanks=False):
    """Checks that a multi-line string is in another and reports a diff output.

    Ignores leading and trailing whitespace on each line, and squeezes repeated
    blank lines down to just one.

    Args:
      ignore_blanks: boolean, whether to ignore blank lines altogether
    """
    exp = ''.join(self._normalize_lines(expected, ignore_blanks=ignore_blanks)).strip()
    act = ''.join(self._normalize_lines(actual, ignore_blanks=ignore_blanks))
    self.assertIn(exp, act, f"""{exp}

not found in:

{act}""")

  @staticmethod
  def _normalize_lines(val, ignore_blanks=False):
    lines = [l.strip() + '\n' for l in val.splitlines(True)]
    return [l for i, l in enumerate(lines)
            if not (ignore_blanks and l == '\n') and
               (i <= 1 or not (lines[i - 1] == l == '\n'))]


class TestCase(mox.MoxTestBase, Asserts):
  """Test case class with lots of extra helpers."""
  maxDiff = None

  def setUp(self):
    # suppress a few warnings
    # local/lib/python3.8/site-packages/mf2util.py:556: DeprecationWarning: The 'warn' function is deprecated, use 'warning' instead
    # logging.warn(f'Failed to parse datetime {date_str}')
    warnings.filterwarnings('ignore', module='mf2util',
                            message="The 'warn' function is deprecated")
    # local/lib/python3.6/site-packages/mox3/mox.py:909: DeprecationWarning: inspect.getargspec() is deprecated, use inspect.signature() or inspect.getfullargspec()
    warnings.filterwarnings('ignore', module='mox', message=r'inspect\.getargspec')
    # local/lib/python3.8/site-packages/webmentiontools/send.py:65: GuessedAtParserWarning: No parser was explicitly specified, so I'm using the best available HTML parser for this system ("lxml"). This usually isn't a problem, but if you run this code on another system, or in a different virtual environment, it may use a different parser and behave differently.
    warnings.filterwarnings('ignore', category=GuessedAtParserWarning)
    # local/lib/python3.9/site-packages/bs4/__init__.py:435: MarkupResemblesLocatorWarning: The input looks more like a filename than markup. You may want to open this file and pass the filehandle into Beautiful Soup.
    warnings.filterwarnings('ignore', category=MarkupResemblesLocatorWarning)
    # local/lib/python3.9/site-packages/bs4/builder/__init__.py:545: XMLParsedAsHTMLWarning: It looks like you're parsing an XML document using an HTML parser. If this really is an HTML document (maybe it's XHTML?)...
    warnings.filterwarnings('ignore', category=XMLParsedAsHTMLWarning)

    super(TestCase, self).setUp()

    appengine_info.APP_ID = 'my-app'
    appengine_info.LOCAL_SERVER = False

    for method in 'get', 'post', 'delete':
      self.mox.StubOutWithMock(requests, method, use_mock_anything=True)
    self.stub_requests_head()

    self.mox.StubOutWithMock(util.urllib.request, 'urlopen')

    # set time zone to UTC so that tests don't depend on local time zone
    os.environ['TZ'] = 'UTC'

    util.follow_redirects_cache.clear()

    util.now = lambda **kwargs: NOW

  def stub_requests_head(self):
    """Automatically return 200 to outgoing HEAD requests."""
    def fake_head(url, **kwargs):
      resp = requests.Response()
      resp.url = url
      if '.' in url or url.startswith('http'):
        resp.headers['content-type'] = 'text/html; charset=UTF-8'
        resp.status_code = 200
      else:
        resp.status_code = 404
      return resp
    self.mox.stubs.Set(requests, 'head', fake_head)

    self._is_head_mocked = False  # expect_requests_head() sets this to True

  def unstub_requests_head(self):
    """Mock outgoing HEAD requests so they must be expected individually."""
    if not self._is_head_mocked:
      self.mox.StubOutWithMock(requests, 'head', use_mock_anything=True)
      self._is_head_mocked = True

  def expect_requests_head(self, *args, **kwargs):
    self.unstub_requests_head()
    return self._expect_requests_call(*args, method=requests.head, **kwargs)

  def expect_requests_get(self, *args, **kwargs):
    return self._expect_requests_call(*args, method=requests.get, **kwargs)

  def expect_requests_post(self, *args, **kwargs):
    return self._expect_requests_call(*args, method=requests.post, **kwargs)

  def expect_requests_delete(self, *args, **kwargs):
    return self._expect_requests_call(*args, method=requests.delete, **kwargs)

  def _expect_requests_call(self, url, response='', status_code=200,
                            content_type='text/html', method=requests.get,
                            redirected_url=None, response_headers=None,
                            **kwargs):
    """
    Args:
      redirected_url: string URL or sequence of string URLs for multiple redirects
    """
    resp = requests_response(
      response, url=url, status=status_code, content_type=content_type,
      redirected_url=redirected_url, headers=response_headers,
      allow_redirects=kwargs.get('allow_redirects'),
      encoding=kwargs.pop('encoding', None))

    if 'timeout' not in kwargs:
      kwargs['timeout'] = HTTP_TIMEOUT
    elif kwargs['timeout'] is None:
      del kwargs['timeout']

    if 'stream' not in kwargs:
      kwargs['stream'] = True
    elif kwargs['stream'] is None:
      del kwargs['stream']

    if method is requests.head:
      kwargs['allow_redirects'] = True

    headers = kwargs.get('headers')
    if not headers:
      headers = kwargs['headers'] = {}

    if not isinstance(headers, mox.Comparator):
      headers.setdefault('User-Agent', util.user_agent)

      def check_headers(actual):
        missing = set(headers.items()) - set(actual.items())
        assert not missing, f'Missing request headers: {missing}\n(Got {set(actual.items())}, expected {set(headers.items())})'
        return True
      kwargs['headers'] = mox.Func(check_headers)

    files = kwargs.get('files')
    if files:
      def check_files(actual):
        self.assertEqual(list(actual.keys()), list(files.keys()))
        for name, expected in files.items():
          self.assertEqual(expected, actual[name].read())
        return True
      kwargs['files'] = mox.Func(check_files)

    call = method(url, **kwargs)
    call.AndReturn(resp)
    return call

  def expect_urlopen(self, url, response=None, status=200, data=None,
                     headers=None, response_headers={}, **kwargs):
    """Stubs out :func:`urllib.request.urlopen()` and sets up an expected call.

    If status isn't 2xx, makes the expected call raise a
    :class:`urllib.error.HTTPError` instead of returning the response.

    If data is set, url *must* be a :class:`urllib.request.Request`.

    If response is unset, returns the expected call.

    Args:
      url: string, :class:`re.RegexObject` or :class:`urllib.request.Request` or
        :class:`webob.request.Request`
      response: string
      status: int, HTTP response code
      data: optional string POST body
      headers: optional expected request header dict
      response_headers: optional response header dict
      kwargs: other keyword args, e.g. timeout
    """
    def check_request(req):
      assert isinstance(req, urllib.request.Request), repr(req)
      try:
        if isinstance(url, RE_TYPE):
          self.assertRegexpMatches(req.get_full_url(), url)
        else:
          self.assertEqual(url, req.get_full_url())

        self.assertEqual(
            data.decode() if isinstance(data, bytes) else data,
            req.data.decode() if isinstance(req.data, bytes) else req.data)

        nonlocal headers
        if isinstance(headers, mox.Comparator):
          self.assertTrue(headers.equals(req.header_items()))
        else:
          if not headers:
            headers = {}
          missing = set(headers.items()) - set(req.header_items())
          assert not missing, f'Missing request headers: {missing}; got {req.header_items()}'

      except AssertionError:
        traceback.print_exc()
        return False

      return True

    if 'timeout' not in kwargs:
      kwargs['timeout'] = HTTP_TIMEOUT

    call = util.urllib.request.urlopen(mox.Func(check_request), **kwargs)
    if status // 100 != 2:
      if response:
        response = urllib.request.addinfourl(io.StringIO(str(response)),
                                             response_headers, url, status)
      call.AndRaise(urllib.error.HTTPError('url', status, 'message',
                                           response_headers, response))
    elif response is not None:
      call.AndReturn(UrlopenResult(status, response, url=url,
                                   headers=response_headers))

    return call



"""Webmention endpoint discovery and sending.

Spec: https://webmention.net/draft/
"""
from collections import namedtuple
import logging
import re
from urllib.parse import urlparse, urljoin

from . import util

logger = logging.getLogger(__name__)

LINK_HEADER_RE = re.compile(
  r'''<([^>]+)>; rel=["']?(https?://)?webmention(\.org/?)?["']?''')

# Returned by discover(). Attributes:
#   endpoint: str
#   response: requests.Response
Endpoint = namedtuple('Endpoint', ('endpoint', 'response'))


def discover(url, follow_meta_refresh=False, **requests_kwargs):
  """Discovers a URL's webmention endpoint.

  Follows up to 30 HTTP 3xx redirects, and at most one client-side HTML meta
  http-equiv=refresh redirects.

  Args:
    url: str
    follow_meta_refresh: bool, whether to follow client side redirects in HTML
      meta http-equiv=refresh tags
    requests_kwargs: passed to :meth:`requests.post`

  Returns: :class:`Endpoint`. If no endpoint is discovered, the endpoint
    attribute will be None.

  Raises: :class:`ValueError` on bad URL, :class:`requests.HTTPError` on failure
  """
  if not url or not isinstance(url, str) or not urlparse(url).netloc:
      raise ValueError(url)

  logger.debug(f'Webmention discovery: attempting for {url}')

  resp = util.requests_get(url, **requests_kwargs)
  # We ignore HTTP status code and allow discovery to continue even on non-2xx
  # responses because the spec doesn't say to stop on error status codes.
  # Background:
  # https://www.w3.org/TR/webmention/#sender-discovers-receiver-webmention-endpoint
  # https://github.com/snarfed/bridgy/issues/1012

  # look in headers
  for link in resp.headers.get('Link', '').split(','):
    match = LINK_HEADER_RE.search(link)
    if match:
      endpoint = util.fragmentless(urljoin(url, match.group(1)))
      logger.debug(f'Webmention discovery: got endpoint in Link header: {endpoint}')
      return Endpoint(endpoint, resp)

  # if no header, require HTML content
  content_type = resp.headers.get('content-type')
  if content_type and content_type.split(';')[0] != 'text/html':
    logger.debug(f'Webmention discovery: no endpoint in headers and content type {content_type} is not HTML')
    return Endpoint(None, resp)

  # look in the content
  soup = util.parse_html(resp.text)
  for tag in soup.find_all(
      ('link', 'a'), attrs={'rel': ('webmention', 'http://webmention.org/')}):
    if tag and tag.get('href'):
      endpoint = util.fragmentless(urljoin(url, tag['href']))
      logger.debug(f'Webmention discovery: got endpoint in tag: {endpoint}')
      return Endpoint(endpoint, resp)

  http_equiv = util.fetch_http_equiv(soup)
  if http_equiv: # else implicit break out and continue like normal
    endpoint = util.fragmentless(urljoin(url, http_equiv))
    if follow_meta_refresh and url != endpoint:
      logger.debug(f'Webmention discovery: following http_equiv: {endpoint}')
      return discover(endpoint, **requests_kwargs)

  logger.debug('Webmention discovery: no endpoint in headers or HTML')
  return Endpoint(None, resp)


def send(endpoint, source, target, **requests_kwargs):
  """Sends a webmention.

  Args:
    endpoint: str, webmention endpoint URL
    source: str, source URL
    target: str, target URL
    requests_kwargs: passed to :meth:`requests.post`

  Returns: :class:`requests.Response` on success.

  Raises: :class:`ValueError` on bad URL, :class:`requests.HTTPError` on failure
  """
  for arg in endpoint, source, target:
    if not arg or not isinstance(arg, str) or not urlparse(arg).netloc:
      raise ValueError(arg)

  logger.debug(f'webmention send: {source} -> {target}')

  requests_kwargs.setdefault('headers', {})['Accept'] = '*/*'
  try:
    # following 3xx redirects translates POST to GET, which we don't want,
    # so disable that. https://github.com/snarfed/bridgy/issues/753
    resp = util.requests_post(endpoint, data={'source': source, 'target': target},
                              allow_redirects=False, **requests_kwargs)
  except BaseException as e:
    logger.debug(f'webmention send: got {e.__class__.__name__}')
    raise

  logger.debug(f'webmention send: got HTTP {resp.status_code} {resp.headers.get("Location", "")}')
  resp.raise_for_status()
  return resp



"""Utility functions for generating Twitter OAuth headers and making API calls.

This is a separate module from twitter.py so that projects like granary can use
it without pulling in App Engine dependencies.

Supports Python 3. Should not depend on App Engine API or SDK packages.
"""
import urllib.request

import requests
import requests_oauthlib
import tweepy

from .webutil import util

TWITTER_APP_KEY = util.read('twitter_app_key')
TWITTER_APP_SECRET = util.read('twitter_app_secret')


def auth_header(url, token_key, token_secret, method='GET'):
  """Generates an Authorization header and returns it in a header dict.

  Args:
    url: string
    token_key: string
    token_secret: string
    method: string

  Returns:
    dict: single element with key 'Authorization'
  """
  oauth1 = requests_oauthlib.OAuth1(
    client_key=TWITTER_APP_KEY,
    client_secret=TWITTER_APP_SECRET,
    resource_owner_key=token_key,
    resource_owner_secret=token_secret,
  )
  req = requests.Request(method=method, url=url, auth=oauth1).prepare()
  return req.headers


def signed_urlopen(url, token_key, token_secret, headers=None, **kwargs):
  """Wraps urllib.request.urlopen() and adds an OAuth signature.
  """
  if headers is None:
    headers = {}

  # if this is a post, move the body params into the URL. Tweepy's OAuth
  # signing doesn't work if they're in the body; Twitter returns a 401.
  data = kwargs.get('data')
  if data:
    method = 'POST'
    url += ('&' if '?' in url else '?') + data
    kwargs['data'] = ''
  else:
    method = 'GET'

  headers.update(auth_header(url, token_key, token_secret, method=method))
  return util.urlopen(urllib.request.Request(url, headers=headers, **kwargs))


def tweepy_auth(token_key, token_secret):
  """Returns a tweepy.OAuth.
  """
  assert TWITTER_APP_KEY and TWITTER_APP_SECRET, \
    "Please fill in the twitter_app_key and twitter_app_secret files in your app's root directory."
  view = tweepy.OAuth(TWITTER_APP_KEY, TWITTER_APP_SECRET)
  view.set_access_token(token_key, token_secret)
  return view



# -*- coding: utf-8 -*-
"""Unit tests for webmention.py."""
import requests

from .. import testutil
from ..webmention import discover, send


class DiscoverTest(testutil.TestCase):

  def _test(self, expected, html, **kwargs):
    call = self.expect_requests_get('http://foo', f'<html>{html}</html>', **kwargs)
    self.mox.ReplayAll()

    got = discover('http://foo')
    self.assertEqual(expected, got.endpoint)
    self.assertEqual(call._return_value, got.response)

  def test_bad_url(self):
    for bad in (None, 123, '', 'asdf'):
      with self.assertRaises(ValueError):
        discover(bad)

  def test_no_endpoint(self):
    self._test(None, '')

  def test_html_link(self):
    self._test('http://endpoint', '<link rel="webmention" href="http://endpoint">')

  def test_html_a(self):
    self._test('http://endpoint', '<a rel="webmention" href="http://endpoint">')

  def test_html_refresh(self):
    call = self.expect_requests_get('http://will/redirect', f'<html><meta http-equiv="refresh" content="0;URL=\'http://foo\'"></html>')
    redirect_call = self.expect_requests_get('http://foo', f'<html><link rel="webmention" href="http://endpoint"></html>')
    self.mox.ReplayAll()

    got = discover('http://will/redirect', follow_meta_refresh=True)
    self.assertEqual('http://endpoint', got.endpoint)
    self.assertEqual(redirect_call._return_value, got.response)

  def test_html_relative(self):
    self._test('http://foo/bar', '<link rel="webmention" href="/bar">')

  def test_html_rel_url(self):
    self._test('http://foo/bar', '<link rel="http://webmention.org/" href="/bar">')

  def test_html_link_and_a(self):
    self._test('http://endpoint1', """\
<a rel="webmention" href="http://endpoint1">
<link rel="webmention" href="http://endpoint2">
""")

  def test_html_link_a_and_refresh(self):
    self._test('http://endpoint1', """\
<meta http-equiv="refresh" content="0;URL=\'http://refresh\'">
<a rel="webmention" href="http://endpoint1">
<link rel="webmention" href="http://endpoint2">
""")

  def test_html_other_links(self):
    self._test('http://endpoint', """\
<link rel="foo" href="http://bar">
<link rel="webmention" href="http://endpoint">
""")

  def test_html_empty(self):
    self._test(None, '<link rel="webmention/" href="">')

  def test_html_unicode(self):
    self._test('http://☕/☕?☕=☕', '<link rel="webmention" href="http://☕/☕?☕=☕">')

  def test_html_escaped(self):
    self._test('http://%E2%98%95/%E2%98%95?%E2%98%95=%E2%98%95',
      '<link rel="webmention" href="http://%E2%98%95/%E2%98%95?%E2%98%95=%E2%98%95">')

  def test_html_content_type_html(self):
    self._test('http://foo/bar', '<link rel="http://webmention.org/" href="/bar">',
               response_headers={'Content-Type': 'text/html'})

  def test_html_content_type_html_charset(self):
    self._test('http://foo/bar', '<link rel="http://webmention.org/" href="/bar">',
               response_headers={'Content-Type': 'text/html; charset=utf-8'})

  def test_html_content_type_other(self):
    self._test(None, '<link rel="http://webmention.org/" href="/bar">',
               response_headers={'Content-Type': 'text/json'})

  def test_header(self):
    self._test('http://endpoint', '', response_headers={
      'Link': '<http://endpoint>; rel=webmention',
    })

  def test_header_relative(self):
    self._test('http://foo/bar', '', response_headers={
      'Link': '</bar>; rel="webmention"',
    })

  def test_header_quoted(self):
    self._test('http://endpoint', '', response_headers={
      'Link': '<http://endpoint>; rel="webmention"',
    })

  def test_header_rel_url(self):
    self._test('http://endpoint', '', response_headers={
      'Link': '<http://endpoint>; rel="https://webmention.org/"',
    })

  def test_other_headers(self):
    self._test('http://endpoint', '', response_headers={
      'Link': '<http://foo>; rel="bar", <http://endpoint>; rel="webmention"',
    })

  def test_multiple_link_headers(self):
    self._test('http://1', '', response_headers={
      'Link': '<http://1>; rel="webmention", <http://2>; rel="webmention"',
    })

  def test_header_empty(self):
    self._test(None, '', response_headers={
      'Link': '<http://endpoint>; rel=""',
    })

  def test_header_query_params(self):
    self._test('http://endpoint?x=y&a=b', '', response_headers={
      'Link': '<http://endpoint?x=y&a=b>; rel="webmention"',
    })

  def test_header_fragment(self):
    self._test('http://endpoint', '', response_headers={
      'Link': '<http://endpoint#foo>; rel="webmention"',
    })

  def test_header_unicode(self):
    self._test('http://☕/☕?☕=☕', '', response_headers={
      'Link': '<http://☕/☕?☕=☕>; rel="webmention"',
    })

  def test_header_escaped(self):
    self._test('http://%E2%98%95/%E2%98%95?%E2%98%95=%E2%98%95', '',
               response_headers={
      'Link': '<http://%E2%98%95/%E2%98%95?%E2%98%95=%E2%98%95>; rel="webmention"',
    })

  def test_http_500(self):
    self._test('http://endpoint', '', status_code=500, response_headers={
      'Link': '<http://endpoint>; rel=webmention',
    })

  def test_connection_error(self):
    self.expect_requests_get('http://foo').AndRaise(
      requests.ConnectionError('foo'))
    self.mox.ReplayAll()

    with self.assertRaises(requests.ConnectionError):
      discover('http://foo')


class SendTest(testutil.TestCase):

  def _test(self, endpoint='http://endpoint', source='http://source',
            target='http://target', **kwargs):
    call = self.expect_requests_post(endpoint, data={
      'source': source,
      'target': target,
    }, allow_redirects=False, headers={'Accept': '*/*'}, **kwargs)
    self.mox.ReplayAll()

    got = send(endpoint, source, target)
    self.assertEqual(call._return_value, got)

  def test_bad_url(self):
    for bad in (None, 123, '', 'asdf'):
      with self.assertRaises(ValueError):
        send(bad, 'http://x', 'http://x')
      with self.assertRaises(ValueError):
        send('http://x', bad, 'http://x')
      with self.assertRaises(ValueError):
        send('http://x', 'http://x', bad)

  def test_success(self):
    self._test()

  def test_requests_exception(self):
    with self.assertRaises(requests.HTTPError):
      self._test(status_code=500)

  def test_preserve_endpoint_query_params(self):
    self._test('http://endpoint?x=y&a=b')

  def test_unicode_urls(self):
    self._test('http://☕/☕?☕=☕', 'http://❤/❤?❤=❤', 'http://⚠/⚠?⚠=⚠')

  def test_unicode_urls_escaped(self):
    self._test('http://%E2%98%95/%E2%98%95?%E2%98%95=%E2%98%95',
               'http://%E2%9A%A0/%E2%9A%A0?%E2%9A%A0=%E2%9A%A0',
               'http://%E2%9D%A4/%E2%9D%A4?%E2%9D%A4=%E2%9D%A4')



"""Utility functions for calling signed Flickr API methods.

Supports Python 3. Should not depend on App Engine API or SDK packages.
"""
import logging
import re
import urllib.error, urllib.parse, urllib.request

import oauthlib.oauth1
import requests_oauthlib
import requests

from .webutil import util
from .webutil.util import json_dumps, json_loads

logger = logging.getLogger(__name__)

FLICKR_APP_KEY = util.read('flickr_app_key')
FLICKR_APP_SECRET = util.read('flickr_app_secret')


def signed_urlopen(url, token_key, token_secret, **kwargs):
  """Call :func:`urllib.request.urlopen`, signing the request with Flickr credentials.

  Args:
    url (string): the url to open
    token_key (string): user's access token
    token_secret (string): the user's access token secret
    timeout (Optional[int]): the request timeout, falls
      back to HTTP_TIMEOUT if not specified

  Returns:
    the file-like object that is the result of :func:`urllib.request.urlopen`
  """
  auth = oauthlib.oauth1.Client(
    FLICKR_APP_KEY,
    client_secret=FLICKR_APP_SECRET,
    resource_owner_key=token_key,
    resource_owner_secret=token_secret)
  uri, headers, body = auth.sign(url, **kwargs)
  try:
    return util.urlopen(urllib.request.Request(uri, body, headers))
  except BaseException as e:
    util.interpret_http_exception(e)
    raise


def raise_for_failure(url, code, msg):
  # https://www.flickr.com/services/api/flickr.auth.checkToken.html#Error%20Codes
  # invalid auth token or API key -> unauthorized
  http_code = 401 if code in (98, 100) else 400
  raise urllib.error.HTTPError(
    url, http_code, f'message={msg}, flickr code={int(code)}', {}, None)


def call_api_method(method, params, token_key, token_secret):
  """Call a Flickr API method.

  Flickr has one API endpoint, where different methods are called by name.

  If the "stat" field contains "fail", then this method creates
  an artificial HTTPError 400 or 401 depending on the type of failure.

  Args:
    method (string): the API method name (e.g. ``flickr.photos.getInfo``)
    params (dict): the parameters to send to the API method
    token_key (string): the user's API access token
    token_secret (string): the user's API access token secret

  Return:
    json object response from the API
  """
  full_params = {
    'nojsoncallback': 1,
    'format': 'json',
    'method': method,
  }
  full_params.update(params)
  url = 'https://api.flickr.com/services/rest?' + urllib.parse.urlencode(full_params)
  resp = signed_urlopen(url, token_key, token_secret)

  text = resp.read()
  try:
    body = json_loads(text)
  except BaseException:
    logger.warning(f'Ignoring malformed flickr response: {text[:1000]}')
    body = {}

  # Flickr returns HTTP success even for errors, so we have to fake it
  if body.get('stat') == 'fail':
    raise_for_failure(url, body.get('code'), body.get('message'))

  return body


def upload(params, file, token_key, token_secret):
  """Upload a photo or video to this user's Flickr account.

  Flickr uploads use their own API endpoint, that returns only XML.
  https://www.flickr.com/services/api/upload.api.html

  Unlike :func:`call_api_method`, this uses the requests library because
  :mod:`urllib` doesn't support multi-part POSTs on its own.

  Args:
    params (dict): the parameters to send to the API method
    file (file-like object): the image or video to upload
    token_key (string): the user's API access token
    token_secret (string): the user's API access token secret

  Return:
    dict containing the photo id (as 'id')

  Raises:
    :class:`requests.HTTPError` on HTTP error or :class:`urllib.error.HTTPError` if
    we get a stat='fail' response from Flickr.
  """
  upload_url = 'https://up.flickr.com/services/upload'
  auth = requests_oauthlib.OAuth1(
      client_key=FLICKR_APP_KEY,
      client_secret=FLICKR_APP_SECRET,
      resource_owner_key=token_key,
      resource_owner_secret=token_secret,
      signature_type=oauthlib.oauth1.SIGNATURE_TYPE_BODY)

  # create a request with files for signing
  faux_req = requests.Request(
    'POST', upload_url, data=params, auth=auth).prepare()
  # parse the signed parameters back out of the body
  data = urllib.parse.parse_qsl(faux_req.body.decode('utf-8'))

  # and use them in the real request
  resp = util.requests_post(upload_url, data=data, files={'photo': file})
  logger.debug(f'upload response: {resp}, {resp.text}')
  resp.raise_for_status()

  m = re.search(r'<rsp stat="(\w+)">', resp.text, re.DOTALL)
  if not m:
    raise BaseException(
      f'Expected response with <rsp stat="...">. Got: {resp.text}')

  stat = m.group(1)
  if stat == 'fail':
    m = re.search(r'<err code="(\d+)" msg="([^"]+)" />', resp.text, re.DOTALL)
    if not m:
      raise BaseException(
        f'Expected response with <err code="..." msg=".." />. Got: {resp.text}')
    raise_for_failure(upload_url, int(m.group(1)), m.group(2))

  m = re.search(r'<photoid>(\d+)</photoid>', resp.text, re.DOTALL)
  if not m:
    raise BaseException(
      f'Expected response with <photoid>...</photoid>. Got: {resp.text}')

  return {'id': m.group(1)}
